
package com.provisionnetworks.rdp5.rdpdr;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Calendar;
import java.util.LinkedList;
import java.util.Locale;
import java.util.TimeZone;

import net.propero.rdp.RdpPacket;
import net.propero.rdp.RdpPacket_Localised;

import org.apache.log4j.Logger;

public class DiskDevice extends Device implements Runnable
{
    protected Logger logger                          = Logger.getLogger( DiskDevice.class );

    protected File                 root                            = null;

    protected FileInfo[]           fileInfo                        = null;

    protected LinkedList           irpQueue                        = null;
    protected Thread               irpThread                       = null;

    protected boolean              isUNIX                          = false;

    protected static final int     MAX_FILES                       = 100;

    protected static final boolean ASYNCHRONOUS                    = false;
    protected static final boolean USE_MONITOR                     = true;

    protected static final int     FS_VOLUME_INFORMATION           = 1;
    protected static final int     FS_LABEL_INFORMATION            = 2;
    protected static final int     FS_SIZE_INFORMATION             = 3;
    protected static final int     FS_DEVICE_INFORMATION           = 4;
    protected static final int     FS_ATTRIBUTE_INFORMATION        = 5;
    protected static final int     FS_CONTROL_INFORMATION          = 6;
    protected static final int     FS_FULLSIZE_INFORMATION         = 7;
    protected static final int     FS_OBJECTID_INFORMATION         = 8;
    protected static final int     FS_MAXIMUM_INFORMATION          = 9;

    protected static final int     FILE_DIRECTORY_INFORMATION      = 1;
    protected static final int     FILE_FULL_DIRECTORY_INFORMATION = 2;
    protected static final int     FILE_BOTH_DIRECTORY_INFORMATION = 3;
    protected static final int     FILE_BASIC_INFORMATION          = 4;
    protected static final int     FILE_STANDARD_INFORMATION       = 5;
    protected static final int     FILE_INTERNAL_INFORMATION       = 6;
    protected static final int     FILE_EA_INFORMATION             = 7;
    protected static final int     FILE_RENAME_INFORMATION         = 10;
    protected static final int     FILE_DISPOSITION_INFORMATION    = 13;
    protected static final int     FILE_END_OF_FILE_INFORMATION    = 20;
    protected static final int     FILE_ATTRIBUTE_TAG_INFORMATION  = 35;

    protected static final int     FILE_NOTIFY_CHANGE_FILE_NAME    = 0x0001;
    protected static final int     FILE_NOTIFY_CHANGE_DIR_NAME     = 0x0002;
    protected static final int     FILE_NOTIFY_CHANGE_ATTRIBUTES   = 0x0004;
    protected static final int     FILE_NOTIFY_CHANGE_SIZE         = 0x0008;
    protected static final int     FILE_NOTIFY_CHANGE_LAST_WRITE   = 0x0010;
    protected static final int     FILE_NOTIFY_CHANGE_SECURITY     = 0x0100;

    protected static final int     FS_CASE_SENSITIVE_SEARCH        = 0x01;
    protected static final int     FS_CASE_PRESERVED_NAMES         = 0x02;

    public DiskDevice( File root )
    {
        super( RdpdrChannel.RDPDR_DTYP_DISK, root.getPath().substring( 0, 1 ) );

        // LogManager.shutdown();

        this.root = root;
        this.fileInfo = new FileInfo[MAX_FILES];
        // this.fileInfo[0] = new FileInfo(root);

        this.isUNIX = !System.getProperty( "os.name" ).contains( "Windows" );

        if ( ASYNCHRONOUS )
        {
            this.irpQueue = new LinkedList();
            this.irpThread = new Thread( this );
            this.irpThread.start();
        }
    }

    public void processIoRequest( IoRequestPacket irp )
    {
        irp.status = RdpdrChannel.STATUS_NOT_IMPLEMENTED;
        irp.status = RdpdrChannel.STATUS_NOT_SUPPORTED;

        switch ( irp.majorFunction )
        {
            case RdpdrChannel.IRP_MJ_CREATE:
                irp.status = this.create( irp );
                break;

            case RdpdrChannel.IRP_MJ_CLOSE:
                irp.status = this.close( irp );
                break;

            case RdpdrChannel.IRP_MJ_READ:
                irp.status = this.read( irp );
                break;

            case RdpdrChannel.IRP_MJ_WRITE:
                irp.status = this.write( irp );
                break;

            case RdpdrChannel.IRP_MJ_QUERY_INFORMATION:
                irp.status = this.queryFileInformation( irp );
                break;

            case RdpdrChannel.IRP_MJ_SET_INFORMATION:
                irp.status = this.setFileInformation( irp );
                break;

            case RdpdrChannel.IRP_MJ_QUERY_VOLUME_INFORMATION:
                irp.status = this.queryVolumeInformation( irp );
                break;

            case RdpdrChannel.IRP_MJ_SET_VOLUME_INFORMATION:
                break;

            case RdpdrChannel.IRP_MJ_DIRECTORY_CONTROL:
            {
                switch ( irp.minorFunction )
                {
                    case RdpdrChannel.IRP_MN_QUERY_DIRECTORY:
                        irp.status = ASYNCHRONOUS ? this
                                .appendIoRequestPacket( irp ) : this
                                .queryDirectory( irp );
                        break;

                    case RdpdrChannel.IRP_MN_NOTIFY_CHANGE_DIRECTORY:
                        irp.status = this.notifyChangeDirectory( irp );
                        break;

                    default:
                        break;
                }

                break;
            }

            case RdpdrChannel.IRP_MJ_FILE_SYSTEM_CONTROL:
                break;

            case RdpdrChannel.IRP_MJ_DEVICE_CONTROL:
                irp.status = this.deviceControl( irp );
                break;

            case RdpdrChannel.IRP_MJ_LOCK_CONTROL:
            case RdpdrChannel.IRP_MJ_QUERY_SECURITY:
            case RdpdrChannel.IRP_MJ_SET_SECURITY:
                break;

            default:
                break;
        }
    }

    private int create( IoRequestPacket irp )
    {
        RdpPacket_Localised packet = irp.packet;

        int desiredAccess = packet.getLittleEndian32();
        long allocationSize = packet.getLittleEndian64();
        int fileAttributes = packet.getLittleEndian32();
        int shareAccess = packet.getLittleEndian32();
        int createDisposition = packet.getLittleEndian32();
        int createOptions = packet.getLittleEndian32();
        int length = packet.getLittleEndian32();
        int information = 0;
        String filename = getUnicodeString( packet, length );

        if ( isUNIX )
        {
            filename = filename.replace( "\\", File.separator );
            filename = filename.replace( "//", File.separator );
        }

        FileInfo fileInfo = null;
        int fileId = 0;

        switch ( createDisposition )
        {
            case RdpdrChannel.FILE_OPEN:
            case RdpdrChannel.FILE_CREATE:
            case RdpdrChannel.FILE_OPEN_IF:
            case RdpdrChannel.FILE_OVERWRITE:
            case RdpdrChannel.FILE_OVERWRITE_IF:
                break;
            default:
                return RdpdrChannel.STATUS_INVALID_PARAMETER;
        }

        File file = new File( this.root, filename );
        if ( file.isDirectory() )
        {
            if ( (createOptions & RdpdrChannel.FILE_NON_DIRECTORY_FILE) != 0 )
            {
                return RdpdrChannel.STATUS_FILE_IS_A_DIRECTORY;
            }
            createOptions |= RdpdrChannel.FILE_DIRECTORY_FILE;
        }

        fileId = allocateFileHandle();

        logger.info( "create - filename=" + file.getPath() + "\tfileId="
                + fileId );
        fileInfo = new FileInfo( file );

        if ( (createOptions & RdpdrChannel.FILE_DIRECTORY_FILE) != 0 )
        {
            switch ( createDisposition )
            {
                case RdpdrChannel.FILE_OPEN:
                    if ( file.isDirectory() )
                        break;
                    return RdpdrChannel.STATUS_NO_SUCH_FILE;
                case RdpdrChannel.FILE_CREATE:
                    file.mkdir();
                    break;
                default:
                    return RdpdrChannel.STATUS_INVALID_PARAMETER;
            }
        }
        else
        {
            if ( file.isDirectory() )
            {
                return RdpdrChannel.STATUS_FILE_IS_A_DIRECTORY;
            }

            logger.info( "createOptions=" + createOptions
                    + "\tcreateDisposition=" + createDisposition
                    + "\tdesireAccess=" + desiredAccess + "\tshareAccess="
                    + shareAccess );

            String mode = "r";

            /*
             * if ((desiredAccess & (RdpdrChannel.GENERIC_ALL |
             * RdpdrChannel.GENERIC_READ | RdpdrChannel.FILE_READ_DATA |
             * RdpdrChannel.FILE_EXECUTE)) != 0) { mode += "r"; }
             */
            if ( (desiredAccess & (RdpdrChannel.GENERIC_ALL
                    | RdpdrChannel.GENERIC_WRITE | RdpdrChannel.FILE_WRITE_DATA)) != 0 )
            {
                mode += "w";
            }

            if ( mode.length() != 0 )
            {
                switch ( createDisposition )
                {
                    case RdpdrChannel.FILE_CREATE:
                        if ( file.exists() )
                            return RdpdrChannel.STATUS_INVALID_PARAMETER;
                        try
                        {
                            fileInfo.raf = new RandomAccessFile( file, mode );
                        }
                        catch( Exception e )
                        {
                            logger.warn( "[DISK] create file with mode: "
                                    + mode + " error: " + e.getMessage() );
                            return RdpdrChannel.STATUS_ACCESS_DENIED;
                        }
                        break;
                    case RdpdrChannel.FILE_OPEN:
                        if ( !file.exists() )
                            return RdpdrChannel.STATUS_NO_SUCH_FILE;
                        try
                        {
                            fileInfo.raf = new RandomAccessFile( file, mode );
                        }
                        catch( Exception e )
                        {
                            logger.warn( "[DISK] open file with mode: " + mode
                                    + " error: " + e.getMessage() );
                            return RdpdrChannel.STATUS_NO_SUCH_FILE;
                        }
                        break;
                    case RdpdrChannel.FILE_OPEN_IF:
                        try
                        {
                            fileInfo.raf = new RandomAccessFile( file, mode );
                            information = 1;
                        }
                        catch( Exception e )
                        {
                            logger.warn( "[DISK] open_if file with mode: "
                                    + mode + " error: " + e.getMessage() );
                            return RdpdrChannel.STATUS_ACCESS_DENIED;
                        }
                        break;
                    case RdpdrChannel.FILE_OVERWRITE:
                    case RdpdrChannel.FILE_OVERWRITE_IF:
                        try
                        {
                            fileInfo.raf = new RandomAccessFile( file, "rw" );
                            fileInfo.raf.setLength( 0 );
                            if ( createDisposition == RdpdrChannel.FILE_OVERWRITE_IF )
                                information = 3;
                        }
                        catch( Exception e )
                        {
                            logger.warn( "[DISK] overwrite file with mode: "
                                    + mode + " error: " + e.getMessage() );
                        }
                        break;
                    default:
                        logger.warn( "[DISK] unknown create disposition: "
                                + createDisposition + " mode: " + mode );
                        return RdpdrChannel.STATUS_ACCESS_DENIED;
                }

            }
            else
            {
                if ( !file.exists() )
                    return RdpdrChannel.STATUS_NO_SUCH_FILE;
            }
        }

        this.fileInfo[fileId] = fileInfo;

        RdpPacket_Localised result = new RdpPacket_Localised( 5 );
        result.setLittleEndian32( fileId ); // fileid
        result.set8( information );

        irp.result = result;

        return RdpdrChannel.STATUS_SUCCESS;
    }

    public int close( IoRequestPacket irp )
    {
        int fileId = irp.fileId;

        if ( !isValidFileHandle( fileId ) )
        {
            return RdpdrChannel.STATUS_INVALID_HANDLE;
        }

        logger.info( "close - fileid=" + fileId );

        FileInfo fileInfo = this.fileInfo[fileId];
        fileInfo.close();
        this.fileInfo[fileId] = null;

        return RdpdrChannel.STATUS_SUCCESS;
    }

    public int read( IoRequestPacket irp )
    {
        int fileId = irp.fileId;
        RdpPacket_Localised packet = irp.packet;

        int length = packet.getLittleEndian32();
        long offset = packet.getLittleEndian64();
        packet.incrementPosition( 20 );

        logger.info( "Read: length=" + length + "\toffset=" + offset );

        if ( !isValidFileHandle( fileId ) )
        {
            return RdpdrChannel.STATUS_INVALID_HANDLE;
        }

        FileInfo fileInfo = this.fileInfo[fileId];

        if ( fileInfo.file.isDirectory() )
        {
            return RdpdrChannel.STATUS_FILE_IS_A_DIRECTORY;
        }

        RandomAccessFile raf = fileInfo.raf;

        try
        {
            raf.seek( offset );
        }
        catch( Exception e )
        {
            return RdpdrChannel.STATUS_ACCESS_DENIED;
        }

        byte[] buffer = new byte[length];
        try
        {
            length = raf.read( buffer );
        }
        catch( Exception e )
        {
            return RdpdrChannel.STATUS_ACCESS_DENIED;
        }

        if ( length > 0 )
        {
            RdpPacket_Localised result = new RdpPacket_Localised( 4 + length );
            result.setLittleEndian32( length );

            result.copyFromByteArray( buffer, 0, result.getPosition(), length );
            result.incrementPosition( length );
            irp.result = result;
        }
        else
        {
            RdpPacket_Localised result = new RdpPacket_Localised( 4 );
            result.setLittleEndian32( 0 );
            irp.result = result;
        }

        return RdpdrChannel.STATUS_SUCCESS;
    }

    public int write( IoRequestPacket irp )
    {
        int fileId = irp.fileId;
        RdpPacket_Localised packet = irp.packet;

        int length = packet.getLittleEndian32();
        long offset = packet.getLittleEndian64();
        packet.incrementPosition( 20 );

        byte[] buffer = null;
        if ( length > 0 )
        {
            buffer = new byte[length];
            packet.copyToByteArray( buffer, 0, packet.getPosition(), length );
            packet.incrementPosition( length );
        }

        if ( !isValidFileHandle( fileId ) )
        {
            return RdpdrChannel.STATUS_INVALID_HANDLE;
        }

        FileInfo fileInfo = this.fileInfo[fileId];
        RandomAccessFile raf = fileInfo.raf;

        try
        {
            raf.seek( offset );
        }
        catch( Exception e )
        {
            logger.warn( "[DISK] Seek exception: " + e.getMessage() );
            return RdpdrChannel.STATUS_ACCESS_DENIED;
        }

        try
        {
            raf.write( buffer );
        }
        catch( Exception e )
        {
            logger.warn( "[DISK] Write exception: " + e.getMessage() );
            return RdpdrChannel.STATUS_ACCESS_DENIED;
        }

        RdpPacket_Localised result = new RdpPacket_Localised( 5 );
        result.setLittleEndian32( buffer.length );
        result.set8( 0 );

        irp.result = result;

        return RdpdrChannel.STATUS_SUCCESS;
    }

    public int queryFileInformation( IoRequestPacket irp )
    {
        int fileId = irp.fileId;
        RdpPacket packet = irp.packet;

        int informationClass = packet.getLittleEndian32();

        int bufferSize = packet.getLittleEndian32();

        logger.info( "FS Information Class=" + informationClass
                + "\tbufferSize=" + bufferSize );
        packet.incrementPosition( 24 );
        packet.incrementPosition( bufferSize );
        if ( !isValidFileHandle( fileId ) )
        {
            return RdpdrChannel.STATUS_INVALID_HANDLE;
        }

        FileInfo fileInfo = this.fileInfo[fileId];
        File file = fileInfo.file;

        irp.result = new RdpPacket_Localised( 4096 );
        irp.result.setLittleEndian32( getFileInformationSize( informationClass,
                file ) );
        int length = fillFileInformation( irp.result, informationClass, file );
        int status = RdpdrChannel.STATUS_SUCCESS;
        if ( length <= 0 )
        {
            status = RdpdrChannel.STATUS_INVALID_PARAMETER;
        }
        return status;
        /*
         * long lastModified = javaTimeToWindowsTime(file.lastModified()); long
         * ftCreationTime = lastModified; long ftLastAccessTime = lastModified;
         * long ftLastWriteTime = lastModified; long ftChangeTime =
         * lastModified; long fileSize = file.length(); long allocationSize =
         * file.length(); int fileAttributes = 0; if (file.isDirectory()) {
         * fileAttributes |= RdpdrChannel.FILE_ATTRIBUTE_DIRECTORY; } if
         * (file.isFile()) { fileAttributes |=
         * RdpdrChannel.FILE_ATTRIBUTE_NORMAL; } if (file.isHidden()) {
         * fileAttributes |= RdpdrChannel.FILE_ATTRIBUTE_HIDDEN; } if
         * (file.canRead() && !file.canWrite()) { fileAttributes |=
         * RdpdrChannel.FILE_ATTRIBUTE_READONLY; } switch (informationClass) {
         * case FILE_BASIC_INFORMATION: result.setLittleEndian32(36); // length
         * result.setLittleEndian64(ftCreationTime); // creation time
         * result.setLittleEndian64(ftLastAccessTime); // last access time
         * result.setLittleEndian64(ftLastWriteTime); // last write time
         * result.setLittleEndian64(ftChangeTime); // change time
         * result.setLittleEndian32(fileAttributes); break; case
         * FILE_STANDARD_INFORMATION: result.setLittleEndian32(22); // length
         * result.setLittleEndian64(allocationSize); // allocation size
         * result.setLittleEndian64(fileSize); // file size
         * result.setLittleEndian32(0); // number of links result.set8(0); //
         * delete pending result.set8((fileAttributes &
         * RdpdrChannel.FILE_ATTRIBUTE_DIRECTORY) != 0 ? 1 : 0); break; case
         * FILE_ATTRIBUTE_TAG_INFORMATION: result.setLittleEndian32(8); //
         * length result.setLittleEndian32(fileAttributes); // file attributes
         * result.setLittleEndian32(0); // reparse tag break; default: return
         * RdpdrChannel.STATUS_NOT_IMPLEMENTED; } return
         * RdpdrChannel.STATUS_SUCCESS;
         */
    }

    public int setFileInformation( IoRequestPacket irp )
    {
        int fileId = irp.fileId;
        RdpPacket_Localised packet = irp.packet;

        int informationClass = packet.getLittleEndian32();
        int bufferSize = packet.getLittleEndian32();
        packet.incrementPosition( 24 );

        if ( !isValidFileHandle( fileId ) )
        {
            return RdpdrChannel.STATUS_INVALID_HANDLE;
        }

        FileInfo fileInfo = this.fileInfo[fileId];
        File file = fileInfo.file;

        switch ( informationClass )
        {

            case FILE_BASIC_INFORMATION:
            {
                long ftCreationTime = packet.getLittleEndian64();
                long ftLastAccessTime = packet.getLittleEndian64();
                long ftLastWriteTime = packet.getLittleEndian64();
                long ftChangeTime = packet.getLittleEndian64();
                int fileAttributes = packet.getLittleEndian32();

                if ( ftLastWriteTime > 0 )
                {
                    file.setLastModified( windowsTimeToJavaTime( ftLastWriteTime ) );
                }
                else if ( ftChangeTime > 0 )
                {
                    file.setLastModified( windowsTimeToJavaTime( ftChangeTime ) );
                }

                if ( fileAttributes != 0 )
                {
                }

                break;
            }

            case FILE_RENAME_INFORMATION:
            {
                boolean replaceIfExists = packet.getLittleEndian16() != 0;
                int filenameLength = packet.getLittleEndian32();
                String filename = getUnicodeString( packet, filenameLength );
                File parent = file.getParentFile();
                if ( isUNIX )
                {
                    filename = filename.replace( "\\", File.separator );
                    filename = filename.replace( "//", File.separator );
                    parent = this.root;
                }

                File newFile = new File( parent, filename );
                file.renameTo( newFile );
                if ( replaceIfExists )
                {
                    try
                    {
                        file.createNewFile();
                    }
                    catch( IOException e )
                    {
                    }
                }
                break;
            }

            case FILE_DISPOSITION_INFORMATION:
            {
                // boolean deleteFile = packet.getLittleEndian16() != 0 ? true :
                // false;
                if ( packet.getPosition() == packet.capacity() )
                {
                    file.delete();
                }

                break;
            }

            default:
                return RdpdrChannel.STATUS_NOT_IMPLEMENTED;
        }

        irp.result = new RdpPacket_Localised( 5 );
        irp.result.setLittleEndian32( bufferSize );
        irp.result.set8( 0 );

        return RdpdrChannel.STATUS_SUCCESS;
    }

    public int queryVolumeInformation( IoRequestPacket irp )
    {
        int fileId = irp.fileId;
        RdpPacket packet = irp.packet;

        int informationClass = packet.getLittleEndian32();
        int bufferSize = packet.getLittleEndian32();

        if ( !isValidFileHandle( fileId ) )
        {
            return RdpdrChannel.STATUS_INVALID_HANDLE;
        }

        FileInfo fileInfo = this.fileInfo[fileId];
        long ftCreationTime = javaTimeToWindowsTime( root.lastModified() );
        int volumeSerialNumber = 0xD838FE64;
        String fileSystemName = "FAT32";

        RdpPacket_Localised result = new RdpPacket_Localised( 4096 );

        switch ( informationClass )
        {
            case FS_VOLUME_INFORMATION:
                result.setLittleEndian32( 17 ); // length
                result.setLittleEndian64( ftCreationTime ); // volume creation
                                                            // time
                result.setLittleEndian32( volumeSerialNumber ); // volume serial
                                                                // number
                result.setLittleEndian32( 0 ); // volume label length
                result.set8( 0 ); // unknown
                break;
            case FS_SIZE_INFORMATION:
                result.setLittleEndian32( 24 ); // length
                result.setLittleEndian64( 0 ); // total allocation units
                result.setLittleEndian64( 0 ); // available allocation units
                result.setLittleEndian32( 0 ); // sectors per allocation unit
                result.setLittleEndian32( 0 ); // bytes per sector
                break;
            case FS_ATTRIBUTE_INFORMATION:
                result.setLittleEndian32( 12 + (fileSystemName.length() + 1) * 2 );
                result.setLittleEndian32( FS_CASE_SENSITIVE_SEARCH
                        | FS_CASE_PRESERVED_NAMES );
                result.setLittleEndian32( 256 );
                result.setLittleEndian32( (fileSystemName.length() + 1) * 2 );
                result.outUnicodeString( fileSystemName,
                        fileSystemName.length() * 2 );
                break;
            default:
                return RdpdrChannel.STATUS_NOT_IMPLEMENTED;
        }

        irp.result = result;

        return RdpdrChannel.STATUS_SUCCESS;
    }

    public int queryDirectory( IoRequestPacket irp )
    {
        int fileId = irp.fileId;
        RdpPacket packet = irp.packet;

        int informationClass = packet.getLittleEndian32();
        int initialQuery = packet.get8();
        int pathLength = packet.getLittleEndian32();
        packet.incrementPosition( 23 );
        String path = getUnicodeString( packet, pathLength );

        if ( isUNIX )
        {
            path = path.replace( "\\", File.separator );
            path = path.replace( "//", File.separator );
        }

        if ( !isValidFileHandle( fileId ) )
        {
            return RdpdrChannel.STATUS_INVALID_HANDLE;
        }

        FileInfo fileInfo = this.fileInfo[fileId];
        int totalLength = 0, lastEntryLength = 0;

        logger.info( "QueryDirectory - " + fileInfo.file.getPath()
                + " - initialQuery=" + initialQuery + ",path=" + path );

        // Execute the query (if needed).
        if ( initialQuery != 0 )
        {
            fileInfo.executeQuery( path );
        }

        // Quit if there are no more results.
        if ( !fileInfo.moreQueryResults() )
        {
            return RdpdrChannel.STATUS_NO_MORE_FILES;
        }

        RdpPacket_Localised result = new RdpPacket_Localised( 4096 );

        // Skip over total length - it will get filled in later.
        result.incrementPosition( 4 );

        // For each directory entry...
        while ( fileInfo.moreQueryResults() )
        {
            // Get the next query result.
            File file = fileInfo.nextQueryResult();

            // Construct the contents of the entry.
            int entryLength = fillFileInformation( result, informationClass,
                    file );
            if ( entryLength < 0 )
            {
                return RdpdrChannel.STATUS_NOT_IMPLEMENTED;
            }
            if ( entryLength == 0 )
                break;

            // Update the total length.
            totalLength += entryLength;

            // Save the last entry length.
            lastEntryLength = entryLength;

            if ( true )
            {
                // TEMPORARY - DO ONE AT A TIME!
                break;
            }
            else
            {
                // Fill in the next entry offset.
                result.setPosition( result.getPosition() - entryLength );
                result.setLittleEndian32( entryLength );
                result.incrementPosition( entryLength - 4 );
            }
        }

        // Fill in the next entry offset of the last entry with a 0.
        result.setPosition( result.getPosition() - lastEntryLength );
        result.setLittleEndian32( 0 );
        result.incrementPosition( lastEntryLength - 4 );

        // Fill in the total length.
        result.setPosition( result.getPosition() - totalLength - 4 );
        result.setLittleEndian32( totalLength );
        result.incrementPosition( totalLength );
        irp.result = result;

        return RdpdrChannel.STATUS_SUCCESS;
    }

    public int notifyChangeDirectory( IoRequestPacket irp )
    {
        int fileId = irp.fileId;
        RdpPacket packet = irp.packet;

        boolean watchSubtree = (packet.get8() != 0 ? true : false);
        int notifyFilter = packet.getLittleEndian32();

        if ( !isValidFileHandle( fileId ) )
        {
            return RdpdrChannel.STATUS_INVALID_HANDLE;
        }

        FileInfo fileInfo = this.fileInfo[fileId];

        if ( USE_MONITOR )
        {
            fileInfo.startMonitor( irp );

            return RdpdrChannel.STATUS_PENDING;
        }
        else
            return RdpdrChannel.STATUS_NOT_IMPLEMENTED;
    }

    public int deviceControl( IoRequestPacket irp )
    {
        /*
         * int fileId = irp.fileId; RdpPacket packet = irp.packet; int
         * outputBufferLength = packet.getLittleEndian32(); int
         * inputBufferLength = packet.getLittleEndian32(); int ioControlCode =
         * packet.getLittleEndian32(); byte[] inputBuffer = null; if
         * (inputBufferLength > 0) { inputBuffer = new byte[inputBufferLength];
         * packet.copyToByteArray(inputBuffer, 0, packet.getPosition(),
         * inputBufferLength); packet.incrementPosition(inputBufferLength); }
         * irp.result = new RdpPacket_Localised(5);
         * irp.result.setLittleEndian32(0); irp.result.set8(0);
         */

        return RdpdrChannel.STATUS_NOT_SUPPORTED;
    }

    // //////////////////////////////////////////////////////////////////
    //
    // File Information Methods
    //
    protected int getFileInformationSize( int informationClass, File file )
    {
        switch ( informationClass )
        {
            case FILE_BOTH_DIRECTORY_INFORMATION:
                return 93 + (file.getName().length() * 2);
            case FILE_BASIC_INFORMATION:
                return 36;
            case FILE_STANDARD_INFORMATION:
                return 22;
            case FILE_ATTRIBUTE_TAG_INFORMATION:
                return 8;
            default:
                break;
        }

        return -1;
    }

    protected int fillFileInformation( RdpPacket_Localised result,
            int informationClass, File file )
    {
        // Determine the required size of the entry.
        int entryLength = getFileInformationSize( informationClass, file );

        // Exit if the information class is not supported.
        if ( entryLength < 0 )
            return -1;

        // Exit if there is no room for the entry.
        if ( entryLength > (result.capacity() - result.getPosition()) )
            return 0;

        // Get directory information for the specified file.
        String fileName = file.getName();
        long lastModified = javaTimeToWindowsTime( file.lastModified() );
        long ftCreationTime = lastModified;
        long ftLastAccessTime = lastModified;
        long ftLastWriteTime = lastModified;
        long ftChangeTime = lastModified;
        long fileSize = file.length();
        long allocationSize = file.length();
        int fileAttributes = 0;

        logger.info( "File size=" + fileSize );

        if ( file.isDirectory() )
        {
            fileAttributes |= RdpdrChannel.FILE_ATTRIBUTE_DIRECTORY;
        }
        if ( fileAttributes == 0 )
        {
            fileAttributes |= RdpdrChannel.FILE_ATTRIBUTE_NORMAL;
        }
        if ( file.isHidden() )
        {
            fileAttributes |= RdpdrChannel.FILE_ATTRIBUTE_HIDDEN;
        }
        if ( file.canRead() && !file.canWrite() )
        {
            fileAttributes |= RdpdrChannel.FILE_ATTRIBUTE_READONLY;
        }

        logger.info( "Fill file information: file name=" + fileName );
        logger.info( "Fill file information: file path=" + file.getPath() );

        // Construct the contents of the entry.
        switch ( informationClass )
        {
            case FILE_BOTH_DIRECTORY_INFORMATION:
                result.setLittleEndian32( entryLength ); // next entry offset
                result.setLittleEndian32( 0 ); // file index
                result.setLittleEndian64( ftCreationTime ); // creation time
                result.setLittleEndian64( ftLastAccessTime ); // last access
                                                              // time
                result.setLittleEndian64( ftLastWriteTime ); // last write time
                result.setLittleEndian64( ftChangeTime ); // change time
                result.setLittleEndian64( fileSize ); // file size
                result.setLittleEndian64( allocationSize ); // allocation size
                result.setLittleEndian32( fileAttributes ); // file attributes

                result.setLittleEndian32( fileName.length() * 2 ); // file name
                                                                   // length
                result.setLittleEndian32( 0 ); // extended attributes length
                result.set8( 0 ); // alternate name (8.3) length
                for( int i = 0; i < 12; i++ )
                {
                    result.setLittleEndian16( 0 ); // alternate name (8.3)
                }
                result.outUnicodeString( fileName, fileName.length() * 2 );
                result.setPosition( result.getPosition() - 2 );

                break;
            case FILE_BASIC_INFORMATION:
                result.setLittleEndian64( ftCreationTime ); // creation time
                result.setLittleEndian64( ftLastAccessTime ); // last access
                                                              // time
                result.setLittleEndian64( ftLastWriteTime ); // last write time
                result.setLittleEndian64( ftChangeTime ); // change time
                result.setLittleEndian32( fileAttributes );
                break;
            case FILE_STANDARD_INFORMATION:
                result.setLittleEndian64( allocationSize ); // allocation size
                result.setLittleEndian64( fileSize ); // file size
                result.setLittleEndian32( 0 ); // number of links
                result.set8( 0 ); // delete pending
                result.set8( (fileAttributes & RdpdrChannel.FILE_ATTRIBUTE_DIRECTORY) != 0 ? 1
                        : 0 );
                break;
            case FILE_ATTRIBUTE_TAG_INFORMATION:
                result.setLittleEndian32( fileAttributes ); // file attributes
                result.setLittleEndian32( 0 ); // reparse tag
                break;
            default:
                logger.info( "Unsupporter feature: FsInformationClass="
                        + informationClass );
                break;
        }

        return entryLength;
    }

    // //////////////////////////////////////////////////////////////////
    //
    // Asynchronous processing
    //
    // //////////////////////////////////////////////////////////////////

    protected int appendIoRequestPacket( IoRequestPacket irp )
    {
        synchronized( irpQueue )
        {
            irpQueue.add( irp );
            irpQueue.notify();
        }
        return RdpdrChannel.STATUS_PENDING;
    }

    public void run()
    {
        for( ;; )
        {
            IoRequestPacket irp = null;

            // Get the next I/O request packet.
            synchronized( irpQueue )
            {
                try
                {
                    while ( irpQueue.isEmpty() )
                    {
                        irpQueue.wait();
                    }
                }
                catch( Exception e )
                {
                    break;
                }
                try
                {
                    irp = (IoRequestPacket) irpQueue.removeFirst();
                }
                catch( Exception e )
                {
                }
            }

            if ( irp == null )
            {
                continue;
            }

            // Process the I/O request packet.
            switch ( irp.majorFunction )
            {
                case RdpdrChannel.IRP_MJ_CREATE:
                    irp.status = this.create( irp );
                    break;

                case RdpdrChannel.IRP_MJ_CLOSE:
                    irp.status = this.close( irp );
                    break;

                case RdpdrChannel.IRP_MJ_READ:
                    irp.status = this.read( irp );
                    break;

                case RdpdrChannel.IRP_MJ_WRITE:
                    irp.status = this.write( irp );
                    break;

                case RdpdrChannel.IRP_MJ_QUERY_INFORMATION:
                    irp.status = this.queryFileInformation( irp );
                    break;

                case RdpdrChannel.IRP_MJ_SET_INFORMATION:
                    irp.status = this.setFileInformation( irp );
                    break;

                case RdpdrChannel.IRP_MJ_QUERY_VOLUME_INFORMATION:
                    irp.status = this.queryVolumeInformation( irp );
                    break;

                case RdpdrChannel.IRP_MJ_SET_VOLUME_INFORMATION:
                    break;

                case RdpdrChannel.IRP_MJ_DIRECTORY_CONTROL:
                {
                    switch ( irp.minorFunction )
                    {
                        case RdpdrChannel.IRP_MN_QUERY_DIRECTORY:
                            irp.status = this.queryDirectory( irp );
                            break;

                        case RdpdrChannel.IRP_MN_NOTIFY_CHANGE_DIRECTORY:
                            irp.status = this.notifyChangeDirectory( irp );
                            break;

                        default:
                            break;
                    }

                    break;
                }

                case RdpdrChannel.IRP_MJ_FILE_SYSTEM_CONTROL:
                    break;

                case RdpdrChannel.IRP_MJ_DEVICE_CONTROL:
                    irp.status = this.deviceControl( irp );
                    break;

                case RdpdrChannel.IRP_MJ_LOCK_CONTROL:
                case RdpdrChannel.IRP_MJ_QUERY_SECURITY:
                case RdpdrChannel.IRP_MJ_SET_SECURITY:
                    break;

                default:
                    break;
            }

            // Send the I/O completion.
            RdpdrChannel.sendIoCompletion( irp );
        }
    }

    // //////////////////////////////////////////////////////////////////
    //
    // Utility Methods
    //
    protected int allocateFileHandle()
    {
        for( int i = 0; i < MAX_FILES; i++ )
        {
            if ( this.fileInfo[i] == null )
                return i;
        }
        return -1;
    }

    protected boolean isValidFileHandle( int fileId )
    {
        return (fileId >= 0) && (fileId < MAX_FILES)
                && (this.fileInfo[fileId] != null) ? true : false;
    }

    protected long windowsTimeToJavaTime( long windowsTime )
    {
        return (windowsTime + getWindowsBaseTime()) / 10000;
    }

    protected long javaTimeToWindowsTime( long javaTime )
    {
        return (javaTime - getWindowsBaseTime()) * 10000;
    }

    protected long getWindowsBaseTime()
    {
        Calendar calendar = Calendar.getInstance(
                TimeZone.getTimeZone( "GMT" ), Locale.getDefault() );
        calendar.set( 1601, 0, 1, 0, 0, 0 );
        long t = calendar.getTimeInMillis();
        return calendar.getTimeInMillis();
    }

    public static String getUnicodeString( RdpPacket packet, int length )
    {
        StringBuffer sb = new StringBuffer();
        for( int i = 0; i < length; i += 2 )
        {
            int unicodeChar = packet.getLittleEndian16();
            if ( unicodeChar == 0 )
                break;
            sb.append( (char) unicodeChar );
        }
        return sb.toString();
    }
}
