package com.provisionnetworks.rdp5.rdpsnd;

import java.io.IOException;

import org.apache.log4j.Logger;

import net.propero.rdp.Options;
import net.propero.rdp.RdesktopException;
import net.propero.rdp.RdpPacket;
import net.propero.rdp.RdpPacket_Localised;
import net.propero.rdp.crypto.CryptoException;
import net.propero.rdp.rdp5.VChannel;

public class RdpsndChannel extends VChannel 
{

	protected Logger logger = Logger.getLogger(RdpsndChannel.class);
	
	protected final static int SNDC_CLOSE 		= 0x01;
	protected final static int SNDC_WAVE		= 0x02;
	protected final static int SNDC_SETVOLUME	= 0x03;
	protected final static int SNDC_SETPITCH	= 0x04;
	protected final static int SNDC_WAVECONFIRM	= 0x05;
	protected final static int SNDC_TRAINING	= 0x06;
	protected final static int SNDC_FORMATS		= 0x07;
	protected final static int SNDC_CRYPTKEY	= 0x08;
	protected final static int SNDC_WAVEENCRYPT	= 0x09;
	
	protected static final int TSSNDCAPS_ALIVE 	= 0x00000001;
    protected static final int TSSNDCAPS_VOLUME = 0x00000002;
    protected static final int TSSNDCAPS_PITCH 	= 0x00000004;
	
    protected static final int PDU_HEADER_SIZE 	= 4;
    
	private WindowsAudioFormat[] audioFormats;
	private int audioFormatsSize = 0;
	
	private int formatNo;
	private int dataSize;
	private byte[] waveData;
	
	private int timeStamp;
	private int blockNo;
	
	public RdpsndChannel( Options options )
	{
	    super( options );
	}
	
	public int flags() 
	{
		return 0;
		//VChannels.CHANNEL_OPTION_INITIALIZED | 
        //VChannels.CHANNEL_OPTION_ENCRYPT_RDP |
        //VChannels.CHANNEL_OPTION_COMPRESS_RDP | 
        //VChannels.CHANNEL_OPTION_SHOW_PROTOCOL;
	}

	public String name() 
	{
		return "RDPSND";
	}

	public void process(RdpPacket packet) throws RdesktopException, IOException,
			CryptoException {
		//Get RDPSND PDU Header
		int msgType = packet.get8();
		packet.incrementPosition(1); // Skip the bPad field
		int bodySize = packet.getLittleEndian16();
		logger.info("[AUDIO] msgType=" + Integer.toHexString(msgType) + ",bodySize=" + Integer.toHexString(bodySize));
		try {
		switch(msgType)
		{
		case 0:
			logger.info("[AUDIO] SNDC_WAVE");
			processWaveData(packet);
			break;
		case SNDC_CLOSE:
			logger.info("[AUDIO] SNDC_CLOSE");
			processSoundClose();
			break;
		case SNDC_WAVE:
			logger.info("[AUDIO] SNDC_WAVEINFO");
			processWaveInfo(packet, bodySize);
			break;
		case SNDC_TRAINING:
			logger.info("[AUDIO] SNDC_TRAINING");
			processTraining(packet);
			break;
		case SNDC_FORMATS:
			logger.info("[AUDIO] SNDC_FORMATS");
			processSoundFormats(packet);
			break;
		}
		}
		catch(Exception e) { logger.error( "[AUDIO] " + e.getMessage(), e ); }
	}
	
	private void processSoundClose()
	{
		// Process the SNDCLOSE PDU
	    try 
        {
            audioFormats[formatNo].stop();
        }
        catch (Exception e) 
        {
            logger.error( "[AUDIO] " + e.getMessage(), e );
        }
	}
	
	private void processWaveInfo(RdpPacket packet, int bodySize)
	{
		// Process the SNDWAVINFO PDU
		final int waveStartDataOffset = 46;
		final int waveStartDataSize = 4;
		final int waveInfoHeaderSize = 8;
		
		timeStamp = packet.getLittleEndian16();
		formatNo = packet.getLittleEndian16();
		blockNo = packet.get8();
		packet.get8();
		packet.getLittleEndian16();
		
		dataSize = bodySize - waveInfoHeaderSize;
		waveData = new byte[dataSize];
		packet.copyToByteArray(waveData, 0, waveStartDataOffset, waveStartDataSize);
	}
	
	private void processWaveData(RdpPacket packet)
	{
		// Process the SNDWAV PDU
		if(dataSize <= 0) return;
		
		final int waveDataOffset = 4;
		packet.copyToByteArray(waveData, waveDataOffset, packet.getPosition(), dataSize - waveDataOffset);
		
		try 
		{
		    if ( audioFormats[formatNo].getState() == WindowsAudioFormat.STATE_UNINITIALIZED )
		    {
		        audioFormats[formatNo] = new WindowsAudioFormat( audioFormats[formatNo] );
		    }
		    
		    if ( audioFormats[formatNo].getPlayState() != WindowsAudioFormat.PLAYSTATE_PLAYING ) 
            {
                audioFormats[formatNo].play();
            }
			audioFormats[formatNo].write( waveData, 0, dataSize );
			audioFormats[formatNo].flush();
		} 
		catch (Exception e) 
		{
			logger.error( "[AUDIO] " + e.getMessage(), e );
		}
		
		sendWaveConfirmPDU();
	}
	
	private void sendWaveConfirmPDU()
	{
		// Send the SNDWAV_CONFIRM PDU
		final int bPad = 0;
		final int packetBodySize = 4;
		final int packetSize = PDU_HEADER_SIZE + packetBodySize;
		
		RdpPacket_Localised packet = new RdpPacket_Localised(packetSize);
		packet.set8(SNDC_WAVECONFIRM);
		packet.set8(bPad);
		packet.setLittleEndian16(packetBodySize);
		packet.setLittleEndian16(timeStamp);
		packet.set8(blockNo);
		packet.set8(bPad);
		packet.markEnd();
		try { send_packet(packet); } catch (Exception e) { logger.error( "[AUDIO] " + e.getMessage(), e ); }
	}
	
	private void processTraining(RdpPacket packet)
	{
		// Process the SNDTRAINING PDU.
    	int wTimeStamp = packet.getLittleEndian16();
    	int wPackSize = packet.getLittleEndian16();
		
    	// Send the SNDTRAININGCONFIRM PDU.
    	final int bPad = 0;
		final int packetBodySize = 4;
		final int packetSize = PDU_HEADER_SIZE + packetBodySize;
		
		RdpPacket_Localised outPacket = new RdpPacket_Localised(packetSize);
		outPacket.set8(SNDC_TRAINING);
		outPacket.set8(bPad);
		outPacket.setLittleEndian16(packetBodySize);
		outPacket.setLittleEndian16(wTimeStamp);
		outPacket.setLittleEndian16(wPackSize);
		outPacket.markEnd();
		try { send_packet(outPacket); } catch (Exception e) { logger.error( "[AUDIO] " + e.getMessage(), e ); }
	}
	
	private void processSoundFormats(RdpPacket packet)
	{
		// Process the SERVER_AUDIO_VERSION_AND_FORMATS PDU
		
		final int sizeOfUnusedValues1 = 14;
		final int sizeOfUnusedValues2 = 3;
		//unused values
		packet.incrementPosition(sizeOfUnusedValues1);//flags, volume, pitch, udp port fields are ignored
		//-----------------------------------------
		
		int numberOfFormats = packet.getLittleEndian16();
		int lastBlockConfirmed = packet.get8();
		
		packet.incrementPosition(sizeOfUnusedValues2);
		
		audioFormats = new WindowsAudioFormat[numberOfFormats];
		audioFormatsSize = 0;
		WindowsAudioFormat audioFormat = null;
		for(int i = 0; i < numberOfFormats; ++i)
		{
		    try
		    {
    			int format = packet.getLittleEndian16();
    			int channels = packet.getLittleEndian16();
    			long samplesPerSec = packet.getLittleEndian32();
    			long avgBytesPerSec = packet.getLittleEndian32();
    			int blockAlign = packet.getLittleEndian16();
    			int bitsPerSample = packet.getLittleEndian16();
    			int dataSize = packet.getLittleEndian16();
    			byte[] data = new byte[dataSize];
    			for(int j = 0; j < dataSize; ++j)
    			{
    				data[j] = (byte)packet.get8();
    			}
    			
    		     logger.info("[AUDIO] New Windos format: format=" + format +
    		                " channels=" + channels + 
    		                " samplePerSec=" + samplesPerSec +
    		                " avgBytesPerSec=" + avgBytesPerSec +
    		                " blockAlign=" + blockAlign +
    		                " bitsPerSample=" + bitsPerSample +
    		                " dataSize=" + dataSize);
    			
    			if( WindowsAudioFormat.getEncoding(format, bitsPerSample) != 0 )
    			{
    				audioFormat = new WindowsAudioFormat(format, channels, samplesPerSec, avgBytesPerSec,
    						blockAlign, bitsPerSample, dataSize, data);
    				if(audioFormat.isSupportedFormat())
    				{
    					audioFormats[audioFormatsSize++] = audioFormat;
    				}
    			}
		    }
		    catch( Exception e )
		    {
		        
		    }
		}
		
		logger.info("[AUDIO] Supported Audio Formats count: " + audioFormatsSize);
		
		// Send client audio formats
		sendClientAudioFormats(lastBlockConfirmed);
	}
	
	private void sendClientAudioFormats(int lastBlockConfirmed)
	{
		// Send the CLIENT_AUDIO_VERSION_AND_FORMATS PDU
		
		int dwFlags = TSSNDCAPS_ALIVE;
		int dwVolume = 0;
		int dwPitch = 0;
		int wDGramPort = 0;
		int wNumberOfFormats = audioFormatsSize;
		int wVersion = 0x05;
		int bPad = 0;
		
		int packetBodySize = 20; // Set packet size - packet info size 
		
		for(int i = 0; i < audioFormatsSize; ++i)
		{
			packetBodySize += audioFormats[i].getSize();
		}
		
		RdpPacket_Localised outPacket = new RdpPacket_Localised(PDU_HEADER_SIZE + packetBodySize);
		
		outPacket.set8(SNDC_FORMATS);
		outPacket.set8(bPad);// Set bPad field
		
		outPacket.setLittleEndian16(packetBodySize);
		outPacket.setLittleEndian32(dwFlags);
		outPacket.setLittleEndian32(dwVolume);//volume ignored
		outPacket.setLittleEndian32(dwPitch);//pitch ignored
		outPacket.setLittleEndian16(wDGramPort);//UDP port ignored
		//-----------------------------------------
		
		outPacket.setLittleEndian16(wNumberOfFormats);
		outPacket.set8(lastBlockConfirmed);
		outPacket.setLittleEndian16(wVersion);//XP protocol version
		outPacket.set8(bPad); // bPad field ignored
		
		for(int i = 0; i < audioFormatsSize; ++i)
		{
			byte[] data = audioFormats[i].toByteArray();
			outPacket.copyFromByteArray(data, 0, outPacket.getPosition(), data.length);
			outPacket.incrementPosition(data.length);
		}
		
		outPacket.markEnd();
		try { send_packet(outPacket); } catch (Exception e) { logger.error(e.getMessage(), e); }
	}

}
