/* Input.java
 * Component: ProperJavaRDP
 * 
 * Revision: $Revision: 1.7 $
 * Author: $Author: telliott $
 * Date: $Date: 2005/09/27 14:15:39 $
 *
 * Copyright (c) 2005 Propero Limited
 *
 * Purpose: Handles input events and sends relevant input data
 *          to server
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * (See gpl.txt for details of the GNU General Public License.)
 * 
 */

package net.propero.rdp;

import java.util.ArrayList;
import java.util.List;

import net.propero.rdp.TouchHandler.TouchMode;

import org.apache.log4j.Logger;

import quest.android.appportal.view.custom.PinchScaling;
import quest.android.appportal.view.custom.ViewDragging;

import android.app.Activity;
import android.graphics.Matrix;
import android.text.Editable;
import android.text.TextWatcher;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnTouchListener;
import android.view.View.OnKeyListener;
import android.widget.EditText;

public abstract class Input implements OnTouchListener 
{
    protected void destroy()
    {
        matrix = null;
        scaling = null;
        dragging = null;
        canvas = null;
        rdp = null;
        frame = null;
        keyAdapter = null;
    }
    
    protected Logger logger = Logger.getLogger(Input.class);

    protected static long last_mousemove = 0;
    
	// Using this flag value (0x0001) seems to do nothing, and after running
	// through other possible values, the RIGHT flag does not appear to be
	// implemented
	protected static final int KBD_FLAG_RIGHT = 0x0001;
	protected static final int KBD_FLAG_EXT = 0x0100;

	// QUIET flag is actually as below (not 0x1000 as in rdesktop)
	protected static final int KBD_FLAG_QUIET = 0x200;
	protected static final int KBD_FLAG_DOWN = 0x4000;
	protected static final int KBD_FLAG_UP = 0x8000;

	protected static final int RDP_KEYPRESS = 0;
	protected static final int RDP_KEYRELEASE = KBD_FLAG_DOWN | KBD_FLAG_UP;
	protected static final int MOUSE_FLAG_MOVE = 0x0800;

	protected static final int MOUSE_FLAG_BUTTON1 = 0x1000;
	protected static final int MOUSE_FLAG_BUTTON2 = 0x2000;
	protected static final int MOUSE_FLAG_BUTTON3 = 0x4000;

	protected static final int MOUSE_FLAG_BUTTON4 = 0x0280; // wheel up -
	                                                        // rdesktop 1.2.0
	protected static final int MOUSE_FLAG_BUTTON5 = 0x0380; // wheel down -
															// rdesktop 1.2.0
	protected static final int MOUSE_FLAG_DOWN = 0x8000;

    protected static final int RDP_INPUT_SYNCHRONIZE = 0;
	protected static final int RDP_INPUT_CODEPOINT = 1;
	protected static final int RDP_INPUT_VIRTKEY = 2;
	protected static final int RDP_INPUT_SCANCODE = 4;
	protected static final int RDP_INPUT_UNICODE = 5;
	protected static final int RDP_INPUT_SCANCODE_TO_UNICODE_FLG = 0x8000;
	protected static final int RDP_INPUT_MOUSE = 0x8001;

	protected int time = 0;

	protected RdesktopCanvas_Localised canvas = null;
	protected Rdp rdp = null;
	protected RdesktopFrame frame = null;
	protected RdesktopKeyAdapter keyAdapter = null;
	
	Matrix matrix = null;
	protected PinchScaling scaling = null;
	protected ViewDragging dragging = null;
	
	protected TouchHandler touchHandler = null;
	
	protected static final int KBD_LWIN_SCANCODE = 0x5B;
	protected static final int KBD_LCTRL_SCANCODE = 0x1D;
	protected static final int KBD_LALT_SCANCODE = 0x38;
	protected static final int KBD_DELETE_SCANCODE = 0x53;
	protected static final int KBD_RIGHT_SCANCODE = 0xCD;
	protected static final int KBD_BACKSPACE_SCANCODE = 0x0E;
	protected static final int KBD_ENTER_SCANCODE = 0x1C;
	protected static final int KBD_TAB_SCANCODE = 0x0F;
	protected static final int KBD_LEFT_ARROW_SCANCODE = 0xCB;
	protected static final int KBD_UP_ARROW_SCANCODE = 0xC8;
	protected static final int KBD_DOWN_ARROW_SCANCODE = 0xD0;
	protected static final int KBD_RIGHT_ARROW_SCANCODE = 0xCD;

	ArrayList<Integer> keycodes = null;
	
    /**
     * Create a new Input object with a given keymap object
     * @param c Canvas on which to listen for input events
     * @param r Rdp layer on which to send input messages
     * @param k Key map to use in handling keyboard events
     */
	public Input( Activity activity, RdesktopCanvas_Localised c, Rdp r, RdesktopFrame f ) 
	{
		canvas = c;
		rdp = r;
		frame = f;
		keyAdapter = new RdesktopKeyAdapter();
		
		matrix = new Matrix();
		scaling = new PinchScaling( activity, frame.desktopView, matrix );
		dragging = new ViewDragging( frame.desktopView, matrix );
		
		touchHandler = new TouchHandler();
		keycodes = new ArrayList<Integer>();
	}

	public void registerCommLayer( Rdp r )
	{
	    this.rdp = r;
	}

	/**
	 * Calls the Rdp.sendInput method.
	 */
	protected void sendInput(final int time, final int message_type, final int device_flags, final int param1, final int param2) 
	{
	    if (!rdp.isConnected()) 
        {
	        logger.debug( "[KEYBOARD] sendInput(): data will not send because RDP is not connected!!!" );
	        return;
        }
	    
	    logger.debug( "[KEYBOARD] sendInput()" );
	    rdp.sendInput(time, message_type, device_flags, param1, param2);
	}
	
	protected void sendInput(int time, int message_type, int device_flags, float param1, float param2) 
	{
	    sendInput( time, message_type, device_flags, (int)param1, (int)param2 );
	}

	public void sendCtrlAltDel()
	{
	    long time = getTime();
	    this.sendScancode( time, KBD_FLAG_DOWN, KBD_LCTRL_SCANCODE );
	    this.sendScancode( time, KBD_FLAG_DOWN, KBD_LALT_SCANCODE );
	    this.sendScancode( time, KBD_FLAG_DOWN, KBD_DELETE_SCANCODE );
	    
	    time = getTime();
        this.sendScancode( time, KBD_FLAG_UP, KBD_LCTRL_SCANCODE );
        this.sendScancode( time, KBD_FLAG_UP, KBD_LALT_SCANCODE );
        this.sendScancode( time, KBD_FLAG_UP, KBD_DELETE_SCANCODE );
	}
	
	public void sendWindowsKey()
	{
	    long time = getTime();
	    this.sendScancode( time, KBD_FLAG_DOWN | KBD_FLAG_EXT, KBD_LWIN_SCANCODE );
	    
	    time = getTime();
	    this.sendScancode( time, KBD_FLAG_UP | KBD_FLAG_EXT, KBD_LWIN_SCANCODE );
	}
	
	/**
     * Retrieve the next "timestamp", by incrementing previous
     * stamp (up to the maximum value of an integer, at which the
     * timestamp is reverted to 1)
     * @return New timestamp value
     */
	public int getTime() 
	{
		time++;
		if (time == Integer.MAX_VALUE)
			time = 1;
		return time;
	}

    /**
     * Send a keyboard event to the server
     * @param time Time stamp to identify this event
     * @param flags Flags defining the nature of the event (eg: press/release/quiet/extended)
     * @param scancode Scancode value identifying the key in question
     */
	public void sendScancode(long time, int flags, int scancode) 
	{
	    logger.debug( "[KEYBOARD] sendScancode(): time=" + time + " flags=" + flags + " scancode=" + scancode );
        
        if ((scancode & RDP_INPUT_SCANCODE_TO_UNICODE_FLG) != 0) 
        {
        	sendInput((int) time, RDP_INPUT_UNICODE, flags, scancode & ~RDP_INPUT_SCANCODE_TO_UNICODE_FLG, 0);
        } 
        else
        {
			sendInput((int) time, RDP_INPUT_SCANCODE, flags, scancode, 0);
        }
	}

    public OnKeyListener getOnKeyListener()
	{
	    return keyAdapter;
	}
	
	public TextWatcher getTextWatcher()
    {
        return keyAdapter;
    }
	
	class RdesktopKeyAdapter implements OnKeyListener, TextWatcher
	{
	    boolean scanned = false;
	    boolean notEmpty = false;
	    
        public boolean onKey( View view, int keyCode, KeyEvent event )
        {
            logger.debug( "[KEYBOARD] Input key event: keycode=" + keyCode + " scanCode=" + event.getScanCode() + " char=" + (char)event.getUnicodeChar() );
            
            this.scanned = true;
            
            int scanCode = event.getScanCode();
            if ( scanCode == 0 )
            {
                switch ( event.getKeyCode() )
                {
                    case KeyEvent.KEYCODE_DEL:
                        try
                        {
                            EditText edit = (EditText) view;
                            if ( edit.getText().length() > 0 )
                            {
                                scanned = false;
                                notEmpty = event.getAction() == KeyEvent.ACTION_DOWN;
                                return false;
                            }
                            else if ( notEmpty )
                            {
                                notEmpty = false;
                                scanned = false;
                                return false;
                            }
                        }
                        catch( Exception e )
                        {
                            
                        }
                        notEmpty = false;
                        scanCode = KBD_BACKSPACE_SCANCODE;
                        break;
                    case KeyEvent.KEYCODE_ENTER:
                    case KeyEvent.KEYCODE_DPAD_CENTER:
                        scanCode = KBD_ENTER_SCANCODE;
                        break;
                    case KeyEvent.KEYCODE_TAB:
                        scanCode = KBD_TAB_SCANCODE;
                        break;
                    default:
                        scanned = false;
                        return false;
                }
            }
            else
            {
                switch ( event.getKeyCode() )
                {
                    case KeyEvent.KEYCODE_DPAD_CENTER:
                        scanCode = KBD_ENTER_SCANCODE;
                        break;
                    case KeyEvent.KEYCODE_DPAD_UP:
                        scanCode = KBD_UP_ARROW_SCANCODE;
                        break;
                    case KeyEvent.KEYCODE_DPAD_DOWN:
                        scanCode = KBD_DOWN_ARROW_SCANCODE;
                        break;
                    case KeyEvent.KEYCODE_DPAD_LEFT:
                        scanCode = KBD_LEFT_ARROW_SCANCODE;
                        break;
                    case KeyEvent.KEYCODE_DPAD_RIGHT:
                        scanCode = KBD_RIGHT_ARROW_SCANCODE;
                        break;
                }
            }
            
            int action = KBD_FLAG_DOWN;
            if ( event.getAction() == KeyEvent.ACTION_UP )
            {
                scanned = false;
                action = KBD_FLAG_UP;
            }
            
            sendScancode( getTime(), action, scanCode );
            
            return false;
        }

        public void onTextChanged(CharSequence s, int start, int before, int count)
        {
            logger.debug( "[KEYBOARD] onTextChanged text: " + s + " [start=" + start + " before=" + before + " count=" + count + "]" );
            
            if ( (s.length() == 0 && before == 0) || scanned ) 
            {
                scanned = false;
                return;
            }
            
            /*
            Integer keyCode = new Integer( s.charAt( s.length() - 1 ) );
            
            logger.debug( "[KEYBOARD] onTextChanged text: " + s + " [start=" + start + " before=" + before + " count=" + count + "]");
            
            int time = getTime();
            sendScancode(time, 0, keyCode.intValue() | RDP_INPUT_SCANCODE_TO_UNICODE_FLG );
            */
            
            String str = s.toString();
            
            String res = count > 0 ? str.substring( start, start + count ) : "";
            
            keycodes.clear();

            int oldSize = str.length() - count + before;
            
            for( int i = start + before; i < oldSize; ++i )
            {
                keycodes.add( KBD_LEFT_ARROW_SCANCODE );
            }
                
            for ( int i = 0; i < before; ++i )
            {
                keycodes.add( KBD_BACKSPACE_SCANCODE );
            }
            
            for ( char c : res.toCharArray() )
            {
                keycodes.add( c | RDP_INPUT_SCANCODE_TO_UNICODE_FLG );
            }
            
            for( int i = start + count; i < str.length(); ++i )
            {
                keycodes.add( KBD_RIGHT_SCANCODE );
            }
            
            sendScancodes( keycodes );
        }
        
        public void afterTextChanged(Editable s) 
        {
            /*
            if ( s.length() == 0 ) return;
            
            else if ( s.length() > 4096 )
            {
                s.delete( 0, s.length() );
            }
            */
        }
        
        public void beforeTextChanged(CharSequence s, int start, int count, int after) { }
	}
	
	private void sendScancodes( List<Integer> keycodes )
	{
	    ArrayList<Rdp.InputEvent> inputList = new ArrayList<Rdp.InputEvent>();
	    
	    for ( int keycode : keycodes )
	    {
	        if ((keycode & RDP_INPUT_SCANCODE_TO_UNICODE_FLG) != 0) 
	        {
	            Rdp.InputEvent keyEvent = new Rdp.InputEvent( time, RDP_INPUT_UNICODE, 0, keycode & ~RDP_INPUT_SCANCODE_TO_UNICODE_FLG, 0 );
	            inputList.add( keyEvent );
	        } 
	        else
	        {
	            Rdp.InputEvent keyDown = new Rdp.InputEvent( time, RDP_INPUT_SCANCODE, KBD_FLAG_DOWN, keycode, 0 );
	            Rdp.InputEvent keyUp = new Rdp.InputEvent( time, RDP_INPUT_SCANCODE, KBD_FLAG_UP, keycode, 0 );
	            inputList.add( keyDown );
                inputList.add( keyUp );
	        }
	    }
	    
	    rdp.sendInput( inputList );
	}
	
	private boolean screenMouseEnabled = false;
	private boolean mouseInMove = false;
	private float mousePosDX = 0;
	private float mousePosDY = 0;
	
	public void setScreenMouseEnabled( boolean enabled, float x, float y )
	{
	    screenMouseEnabled = enabled;
	    canvas.movePointer( (int)x, (int)y );
	    canvas.setPointerEnabled( enabled );
	}
	
    public boolean isScreenMouseEnabled( )
    {
        return screenMouseEnabled;
    }
    
    public boolean onTouch( final View v, final MotionEvent event )
    {
        try
        {
            return onTouchThread( v, event );
        }
        catch( Throwable e )
        {
            logger.error( "Cannot send RDP input touch: " + e.getMessage(), e );
        }
        
        return true;
    }
    
    private void sendMouseLeftClick( int time, int x, int y )
    {
        ArrayList<Rdp.InputEvent> inputList = new ArrayList<Rdp.InputEvent>(2);
        
        Rdp.InputEvent press = new Rdp.InputEvent( time, RDP_INPUT_MOUSE, MOUSE_FLAG_BUTTON1 | MOUSE_FLAG_DOWN, x, y );
        Rdp.InputEvent release = new Rdp.InputEvent( time, RDP_INPUT_MOUSE, MOUSE_FLAG_BUTTON1, x, y );

        inputList.add( press );
        inputList.add( release );
        
        rdp.sendInput( inputList );
    }
    
    private void sendMouseRightClick( int time, int x, int y )
    {
        ArrayList<Rdp.InputEvent> inputList = new ArrayList<Rdp.InputEvent>(2);
        
        Rdp.InputEvent press = new Rdp.InputEvent( time, RDP_INPUT_MOUSE, MOUSE_FLAG_BUTTON2 | MOUSE_FLAG_DOWN, x, y );
        Rdp.InputEvent release = new Rdp.InputEvent( time, RDP_INPUT_MOUSE, MOUSE_FLAG_BUTTON2, x, y );

        inputList.add( press );
        inputList.add( release );
        
        rdp.sendInput( inputList );
    }
    
    long lastTimeClick = 0;
    float prevFastTouchX = 0;
    float prevFastTouchY = 0;
    
    public boolean onTouchThread( View v, MotionEvent event )
    {
        float x = event.getX();
        float y = event.getY();
        
        float[] values = new float[9];
        matrix.getValues( values );

        float scale = values[Matrix.MSCALE_X];
        
        float xPos = values[Matrix.MTRANS_X];
        float yPos = values[Matrix.MTRANS_Y];
        
        x = (x - xPos) / scale;
        y = (y - yPos) / scale;
        
        TouchMode mode = touchHandler.handleTouch( event );
        
        if ( screenMouseEnabled )
        {
            processTouchWithMouse( mode, event, x, y );
        }
        else
        {
            processTouchWithoutMouse( mode, event, x, y );
        }
        
        return true;
    }
    
    private void processTouchWithMouse( TouchMode mode, MotionEvent event, float x, float y )
    {
        boolean isPointerTouched = mouseInMove || canvas.isPointerTouched( x, y );
        
        if ( isPointerTouched )
        {
            switch( mode )
            {
                case START_SINGLE_TOUCH:
                    mousePosDX = canvas.getPointerPosX() - x;
                    mousePosDY = canvas.getPointerPosY() - y;
                    canvas.setTouched( true );
                    mouseInMove = true;
                    break;
                    
                case END_SINGLE_TOUCH:
                case SINGLE_END_MOVE:
                    canvas.setTouched( false );
                    mouseInMove = false;
                    break;
                    
                case SINGLE_FAST_TOUCH:
                    mouseInMove = false;
                    
                    // Calculate mouse position
                    if ( mousePosDX > 0 ) mousePosDX = 0;
                    if ( mousePosDY > 0 ) mousePosDY = 0;
                    canvas.movePointer( (int)(x + mousePosDX), (int)(y + mousePosDY) );
                    x = canvas.getPointerX();
                    y = canvas.getPointerY();
                    
                    if ( System.currentTimeMillis() - lastTimeClick < 500 )
                    {
                        float r2 = (x - prevFastTouchX) * (x - prevFastTouchX) +
                        (y - prevFastTouchY) * (y - prevFastTouchY);
                            
                        if ( r2 < 225 )
                        {
                            sendMouseLeftClick( time, (int)x, (int)y );
                        }
                    }
                    else
                    {
                        prevFastTouchX = x;
                        prevFastTouchY = y;
                        lastTimeClick = System.currentTimeMillis();
                    }
                    
                    sendMouseLeftClick( time, (int)x, (int)y );
                    canvas.setTouched( false );
                    break;
                    
                case SINGLE_LONG_TOUCH:
                    mouseInMove = false;
                    if ( mousePosDX > 0 ) mousePosDX = 0;
                    if ( mousePosDY > 0 ) mousePosDY = 0;
                    canvas.movePointer( (int)(x + mousePosDX), (int)(y + mousePosDY) );
                    x = canvas.getPointerX();
                    y = canvas.getPointerY();
                    
                    sendMouseRightClick( time, (int)x, (int)y );
                    canvas.setTouched( false );
                    break;
                    
                case SINGLE_START_MOVE:
                    mouseInMove = true;
                    canvas.movePointer( (int)(x + mousePosDX), (int)(y + mousePosDY) );
                    x = canvas.getPointerX();
                    y = canvas.getPointerY();
                    sendInput( time, RDP_INPUT_MOUSE, MOUSE_FLAG_MOVE, x, y );
                    break;
                    
                case SINGLE_MOVE:
                case SINGLE_MOVE_AFTER_LONG_TOUCH:
                    canvas.movePointer( (int)(x + mousePosDX), (int)(y + mousePosDY) );
                    x = canvas.getPointerX();
                    y = canvas.getPointerY();
                    sendInput( time, RDP_INPUT_MOUSE, MOUSE_FLAG_MOVE, x, y );
                    break;
                    
                case SINGLE_START_MOVE_AFTER_LONG_TOUCH:
                    float posX = canvas.getPointerX();
                    float posY = canvas.getPointerY();
                    sendInput( time, RDP_INPUT_MOUSE, MOUSE_FLAG_BUTTON1 | MOUSE_FLAG_DOWN, posX, posY );
                    canvas.movePointer( (int)(x + mousePosDX), (int)(y + mousePosDY) );
                    x = canvas.getPointerX();
                    y = canvas.getPointerY();
                    sendInput( time, RDP_INPUT_MOUSE, MOUSE_FLAG_MOVE, x, y );
                    break;
                
                case SINGLE_END_MOVE_AFTER_LONG_TOUCH:
                    mouseInMove = false;
                    canvas.movePointer( (int)(x + mousePosDX), (int)(y + mousePosDY) );
                    x = canvas.getPointerX();
                    y = canvas.getPointerY();
                    sendInput( time, RDP_INPUT_MOUSE, MOUSE_FLAG_BUTTON1, x, y );
                    canvas.setTouched( false );
                    break;
                    
                case MULTIPLE_FAST_TOUCH:
                    x = canvas.getPointerX();
                    y = canvas.getPointerY();
                    sendMouseRightClick( time, (int)x, (int)y );
                    break;
                    
                // case MULTIPLE_MOVE_TOUCH:
                //case MULTIPLE_WAIT:
                //case START_MULTIPLE_TOUCH:
                case END_MULTIPLE_TOUCH:
                    canvas.setTouched( false );
                    mouseInMove = false;
                    break;
                    
                default:
                    break;
            }
        }
        else
        {
            processTouchWithoutMouse( mode, event, x, y );
        }
    }
    
    private void processTouchWithoutMouse( TouchMode mode, MotionEvent event, float x, float y )
    {
        switch( mode )
        {
            case START_SINGLE_TOUCH:
            case END_SINGLE_TOUCH:
            default:
                break;
                
            case SINGLE_FAST_TOUCH:
                if ( System.currentTimeMillis() - lastTimeClick < 500 )
                {
                    float r2 = (x - prevFastTouchX) * (x - prevFastTouchX) +
                    (y - prevFastTouchY) * (y - prevFastTouchY);
                        
                    if ( r2 < 225 )
                    {
                        sendMouseLeftClick( time, (int)x, (int)y );
                    }
                }
                else
                {
                    prevFastTouchX = x;
                    prevFastTouchY = y;
                    lastTimeClick = System.currentTimeMillis();
                }
                
                sendMouseLeftClick( time, (int)x, (int)y );
                break;
                
            case SINGLE_LONG_TOUCH:
                sendMouseRightClick( time, (int)x, (int)y );
                break;
                
            case SINGLE_START_MOVE:
                dragging.startDragging( event );
                break;
                
            case SINGLE_MOVE:
            case SINGLE_END_MOVE:
                dragging.processDragging( event );
                break;
                
            case SINGLE_START_MOVE_AFTER_LONG_TOUCH:
                float oldX = x;
                float oldY = y;
                
                int historySize = event.getHistorySize(); 
                
                if ( historySize > 1 )
                {
                    oldX = event.getHistoricalX( historySize - 1 );
                    oldY = event.getHistoricalY( historySize - 1 );
                }
                
                sendInput( time, RDP_INPUT_MOUSE, MOUSE_FLAG_BUTTON1 | MOUSE_FLAG_DOWN, oldX, oldY );
                sendInput( time, RDP_INPUT_MOUSE, MOUSE_FLAG_MOVE, x, y );
                break;
                
            case SINGLE_MOVE_AFTER_LONG_TOUCH:
                sendInput( time, RDP_INPUT_MOUSE, MOUSE_FLAG_MOVE, x, y );
                break;
                
            case SINGLE_END_MOVE_AFTER_LONG_TOUCH:
                sendInput( time, RDP_INPUT_MOUSE, MOUSE_FLAG_BUTTON1, x, y );
                break;
                
            case START_MULTIPLE_TOUCH:
                scaling.startScaling( event );
                break;
                
            case MULTIPLE_MOVE_TOUCH:
                scaling.processScaling( event );
                break;
                
            case END_MULTIPLE_TOUCH:
                break;
                
            case MULTIPLE_FAST_TOUCH:
                sendMouseRightClick( time, (int)x, (int)y );
                break;
        }
    }
}
