package net.propero.rdp;

import java.util.Hashtable;

import org.apache.log4j.Logger;

import android.graphics.PointF;
import android.util.Log;
import android.view.MotionEvent;

public class TouchHandler
{
    Logger logger = Logger.getLogger(TouchHandler.class);
    
    public enum TouchMode
    {
        NO_TOUCH,
        
        START_SINGLE_TOUCH,
        SINGLE_WAIT,
        SINGLE_FAST_TOUCH,
        SINGLE_LONG_TOUCH,
        SINGLE_START_MOVE,
        SINGLE_MOVE,
        SINGLE_END_MOVE,
        SINGLE_START_MOVE_AFTER_LONG_TOUCH,
        SINGLE_MOVE_AFTER_LONG_TOUCH,
        SINGLE_END_MOVE_AFTER_LONG_TOUCH,
        END_SINGLE_TOUCH,
        
        START_MULTIPLE_TOUCH,
        MULTIPLE_WAIT,
        MULTIPLE_FAST_TOUCH,
        MULTIPLE_MOVE_TOUCH,
        END_MULTIPLE_TOUCH;
    }
    
    private static final int ONE_TOUCH = 1;
    
    private static final int FIRST_FINGER_IDX = 0;
    private static final int SECOND_FINGER_IDX = 1;
    
    private static final float TOUCH_RADIUS_THRESHOLD = 1F; // pixels
    private static final float TOUCH_RADIUS_THRESHOLD2 = TOUCH_RADIUS_THRESHOLD * TOUCH_RADIUS_THRESHOLD;
    
    private static final int FAST_TOUCH_THRESHOLD = 1000;
    
    TouchMode mode = TouchMode.NO_TOUCH;
    
    boolean isMoved = false;
    
    Hashtable<Integer, PointF> downPoint = new Hashtable<Integer, PointF>();
    Hashtable<Integer, Long> downTime = new Hashtable<Integer, Long>();
    
    public TouchMode handleTouch( MotionEvent event )
    {
        mode = process( event, mode );
        
        logger.debug( "[TOUCH] " + mode.toString() );
        
        return mode;
    }
    
    private boolean isMoved( MotionEvent event, int pointerIdx )
    {
        float x1 = downPoint.get( pointerIdx ).x;
        float y1 = downPoint.get( pointerIdx ).y;
        
        float x2 = event.getX( pointerIdx );
        float y2 = event.getY( pointerIdx );
        
        float r = (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1);
        
        if ( r > TOUCH_RADIUS_THRESHOLD2 ) 
        {
            return true;
        }
        return false;
    }
    
    private TouchMode process( MotionEvent event, TouchMode currentMode )
    {
        int action = event.getActionMasked();
        
        boolean isDown = action == MotionEvent.ACTION_DOWN || action == MotionEvent.ACTION_POINTER_DOWN;
        
        boolean isUp = action == MotionEvent.ACTION_UP || 
        action == MotionEvent.ACTION_CANCEL || 
        action == MotionEvent.ACTION_POINTER_UP;
        
        boolean isMove = action == MotionEvent.ACTION_MOVE;
        
        boolean isMoveAccepted = isMoved || (isMove && isMoved( event, event.getActionIndex() ));
        isMoved = isMoveAccepted;
        
        boolean isMultipleTouch = event.getPointerCount() > ONE_TOUCH;
        
        long time = downTime.containsKey( event.getActionIndex() ) ? downTime.get( event.getActionIndex() ) : event.getDownTime();
        boolean isFastTouch = (event.getEventTime() - time < FAST_TOUCH_THRESHOLD);
        
        boolean isSecondFinger = event.getActionIndex() == SECOND_FINGER_IDX;
        
        boolean isSingleMoveStarted = currentMode == TouchMode.SINGLE_START_MOVE;
        boolean isSingleMoveInProcess = currentMode == TouchMode.SINGLE_MOVE;
        boolean isSingleMoveAfterLongTouchStarted = currentMode == TouchMode.SINGLE_START_MOVE_AFTER_LONG_TOUCH;
        boolean isSingleMoveAfterLongTouchInProcess = currentMode == TouchMode.SINGLE_MOVE_AFTER_LONG_TOUCH;
        
        boolean isMultiTouchEnded = currentMode == TouchMode.END_MULTIPLE_TOUCH;
        boolean isMultipleFastTouch = currentMode == TouchMode.MULTIPLE_FAST_TOUCH;
        
        if ( isDown )
        {
            downPoint.put( event.getActionIndex(), new PointF( event.getX(), event.getY() ) );
            downTime.put( event.getActionIndex(), event.getEventTime() );
        }
        else if ( isUp )
        {
            downPoint.remove( event.getActionIndex() );
            downTime.remove( event.getActionIndex() );
        }
        
        if ( isUp )
        {
            isMoved = false;
        }
        
        if ( isDown && isSecondFinger )
        {
            isMoved = false;
            
            downPoint.remove( FIRST_FINGER_IDX );
            downTime.remove( FIRST_FINGER_IDX );
            
            downPoint.put( FIRST_FINGER_IDX, new PointF( event.getX(FIRST_FINGER_IDX), event.getY(FIRST_FINGER_IDX) ) );
            downTime.put( FIRST_FINGER_IDX, event.getEventTime() );
        }
        
        if ( !isMultipleTouch && !isDown )
        {
            if ( isMultiTouchEnded )
            {
                return TouchMode.END_MULTIPLE_TOUCH;
            }
            else if ( isMultipleFastTouch )
            {
                isMoved = true;
                return TouchMode.SINGLE_START_MOVE;
            }
        }
            
        switch( event.getActionMasked() )
        {
            case MotionEvent.ACTION_DOWN:
            case MotionEvent.ACTION_POINTER_DOWN:
                if ( !isMultipleTouch )
                {
                    return TouchMode.START_SINGLE_TOUCH;
                }
                else
                {
                    return TouchMode.START_MULTIPLE_TOUCH;
                }
                
            case MotionEvent.ACTION_MOVE:
                if ( !isMoveAccepted )
                {
                    if ( isMultipleTouch )
                    {
                        return TouchMode.MULTIPLE_WAIT;
                    }
                    else
                    {
                        return TouchMode.SINGLE_WAIT;
                    }
                }
                else
                {
                    if ( isMultipleTouch )
                    {
                        return TouchMode.MULTIPLE_MOVE_TOUCH;
                    }
                    else
                    {
                        if ( isFastTouch && !isSingleMoveInProcess && !isSingleMoveStarted )
                        {
                            return TouchMode.SINGLE_START_MOVE;
                        }
                        else
                        {
                            if ( isSingleMoveInProcess || isSingleMoveStarted )
                            {
                                return TouchMode.SINGLE_MOVE;
                            }
                            else if ( !isSingleMoveAfterLongTouchInProcess && !isSingleMoveAfterLongTouchStarted )
                            {
                                return TouchMode.SINGLE_START_MOVE_AFTER_LONG_TOUCH;
                            }
                            else
                            {
                                return TouchMode.SINGLE_MOVE_AFTER_LONG_TOUCH;
                            }
                        }
                    }
                }
                    
            case MotionEvent.ACTION_CANCEL:
            case MotionEvent.ACTION_UP:
            case MotionEvent.ACTION_POINTER_UP:
                if ( isMultipleTouch )
                {
                    if ( isMoveAccepted )
                    {
                        return TouchMode.END_MULTIPLE_TOUCH;
                    }
                    else
                    {
                        if ( isSecondFinger )
                        {
                            if ( isFastTouch )
                            {
                                return TouchMode.MULTIPLE_FAST_TOUCH;
                            }
                            else
                            {
                                return TouchMode.END_MULTIPLE_TOUCH;
                            }
                        }
                    }
                }
                else 
                {
                    if ( !isMoveAccepted )
                    {
                        if ( isFastTouch )
                        {
                            return TouchMode.SINGLE_FAST_TOUCH;
                        }
                        else
                        {
                            return TouchMode.SINGLE_LONG_TOUCH;
                        }
                    }
                    else
                    {
                        if ( isSingleMoveStarted || isSingleMoveInProcess )
                        {
                            return TouchMode.SINGLE_END_MOVE;
                        }
                        else if ( isSingleMoveAfterLongTouchStarted || isSingleMoveAfterLongTouchInProcess )
                        {
                            return TouchMode.SINGLE_END_MOVE_AFTER_LONG_TOUCH;
                        }
                    }
                }
                break;
        }
        
        return mode;
    }
}
