package net.propero.rdp.orders;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.HashMap;

import org.apache.log4j.Logger;

import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Path;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Region.Op;

import com.provisionnetworks.rdp5.pnrdpp.PnrdppChannel;

import net.propero.rdp.RdesktopCanvas;
import net.propero.rdp.RdpPacket_Localised;
import net.propero.rdp.orders.emfplus.*;
import net.propero.rdp.orders.emfplus.enumerations.*;

public class AlternateSecondaryOrders {

	Logger logger = Logger.getLogger(AlternateSecondaryOrders.class);
	
	public static final int TS_ALTSEC_SWITCH_SURFACE = 0x00;
	public static final int TS_ALTSEC_CREATE_OFFSCR_BITMAP = 0x01;
	public static final int TS_ALTSEC_STREAM_BITMAP_FIRST = 0x02;
	public static final int TS_ALTSEC_STREAM_BITMAP_NEXT = 0x03;
	public static final int TS_ALTSEC_CREATE_NINEGRID_BITMAP = 0x04;
	public static final int TS_ALTSEC_GDIP_FIRST = 0x05;
	public static final int TS_ALTSEC_GDIP_NEXT = 0x06;
	public static final int TS_ALTSEC_GDIP_END = 0x07;
	public static final int TS_ALTSEC_GDIP_CACHE_FIRST = 0x08;
	public static final int TS_ALTSEC_GDIP_CACHE_NEXT = 0x09;
	public static final int TS_ALTSEC_GDIP_CACHE_END = 0x0a;
	public static final int TS_ALTSEC_WINDOW = 0x0b;
	public static final int TS_ALTSEC_COMPDESK_FIRST = 0x0c;
	
	public static final String PNRDP_EMFP_COMMENT_STRING = "PNRDP";
	
	private int totalSize = 0;
	private int orderBufferSize = 0;
	private byte[] orderBuffer = null;
	
	private HashMap<Integer,EmfPlusRecord> objectTable = new HashMap<Integer,EmfPlusRecord>();
	
	public void processAltSecOrder(RdpPacket_Localised data, int order_flags, RdesktopCanvas surface)
	{
		logger.info("Process Alternate Secondary Orders");
    	int type = (order_flags >> 2); 
        data.get8(); // flags - not used
        int size = data.getLittleEndian16();
        switch(type)
        {
        case TS_ALTSEC_GDIP_FIRST:
        	totalSize = data.getLittleEndian32();
        	data.getLittleEndian32(); // total EMF size - not used
        	
        	orderBuffer = new byte[totalSize];
        	orderBufferSize = size;
        	data.copyToByteArray(orderBuffer, 0, data.getPosition(), size);        	
        	data.incrementPosition(size);
        	break;
        case TS_ALTSEC_GDIP_NEXT:
        	
        	data.copyToByteArray(orderBuffer, orderBufferSize, data.getPosition(), size);   
        	data.incrementPosition(size);
        	
        	orderBufferSize += size;
        	break;
        case TS_ALTSEC_GDIP_END:
        	totalSize = data.getLittleEndian32();
        	data.getLittleEndian32(); // total EMF size - not used
        	
        	data.copyToByteArray(orderBuffer, orderBufferSize, data.getPosition(), size); 
        	data.incrementPosition(size);
        	
        	orderBufferSize = totalSize;
        	break;
        default:
        	logger.error("Incorrect Alternate Order: " + type);
        	break;
        }
        
        if (orderBufferSize == totalSize)
        {
            // process the orders
            if (!processOrder(orderBuffer, orderBufferSize, surface))
            {
                logger.error("Process order failed");
            }
            
        }
	}
		
	private boolean processOrder(byte[] data, int size, RdesktopCanvas surface) {
		logger.info("Process Orders");
		try {
			OrderBuffer buffer = new OrderBuffer(size);
			buffer.copyFromByteArray(data, 0, 0, size);
			
			Canvas canvas = surface.getInnerGraphics();
			
			while(buffer.getPosition() < size)
			{	
				EmfPlusRecord record = EmfPlusObjectsFactory.getRecord(buffer);
				if(record == null) {
					logger.error("Error getting record");
					return false;
				}
				switch(record.getType()) {
				case RecordTypeEnumeration.EmfPlusComment:
					if (!checkComment(record)) {
						return false;
					}
					break;
				case RecordTypeEnumeration.EmfPlusObject:
					EmfPlusObject object = (EmfPlusObject)record;
					objectTable.put(object.getObjectID(), object);
					break;
				case RecordTypeEnumeration.EmfPlusDrawImage:
					EmfPlusDrawImage drawImage = (EmfPlusDrawImage)record;
					EmfPlusImage emfImage = (EmfPlusImage)objectTable.get(drawImage.getObjectID());
					drawImage(drawImage, emfImage, canvas);
					
					surface.repaint((int)drawImage.getDestRect().x, 
								(int)drawImage.getDestRect().y, 
								(int)drawImage.getDestRect().width, 
								(int)drawImage.getDestRect().height);
					
					break;
				case RecordTypeEnumeration.EmfPlusSetClipRegion:
					EmfPlusSetClipRegion setClip = (EmfPlusSetClipRegion)record;
					EmfPlusRegion region = (EmfPlusRegion)objectTable.get(setClip.getObjectID());
					setClippingRegion(setClip, region, canvas);
					break;
				default:
					logger.error("Object type is unknown: " + Integer.toHexString(record.getType()));
					return false;
				}
			}
			//graphics.dispose();
		}
		catch(Exception e) {
			logger.error("Error in process orders: " + e.getMessage());
			return false;
		}
		
		return true;
	}

	private boolean checkComment(EmfPlusRecord record) {
		logger.info("Check Comment");
		EmfPlusComment comment = (EmfPlusComment)record;
		try
		{
			String commentString = new String(comment.getCommentData());
			logger.info("Comment String: " + commentString);
			if (commentString.startsWith(PNRDP_EMFP_COMMENT_STRING))
			{
				String versionString = commentString.substring(PNRDP_EMFP_COMMENT_STRING.length());
				logger.info("Server PNRDP Version: " + versionString);
				if ( !versionString.equals(PnrdppChannel.ProtocolVersion + "") )
				{
					logger.error("Incorrect comment: protocol version incorrect: " + versionString);
					return false;
				}
			}
			else
			{
				logger.error("Incorrect comment: " + comment);
				return false;
			}
		}
		catch(Exception e)
		{
			logger.error("Error in Comment Parsing: " + e.getMessage());
			return false;
		}
		return true;
	}
	
	private void setClippingRegion(EmfPlusSetClipRegion record, EmfPlusRegion region, /*Graphics*/Canvas graphics)
	{
		logger.info("Set Clip Region");
		if (record.getCombineMode() != CombineModeEnumeration.CombineModeReplace)
		{
			logger.info("Warning: Different combine mode - " + Integer.toHexString(record.getCombineMode()));
		}
		
		for(int i = 0; i < region.getNodes().length; ++i)
		{
			EmfPlusRegionNode node = region.getNodes()[i];
			switch(node.getType())
			{
			case RegionNodeDataTypeEnumeration.RegionNodeDataTypePath:
				EmfPlusPath path = node.getRegionNodePath().getPath();
	            if (path.getPbit())
	            {
	            	logger.info("Locations are specified by EmfPlusPointR objects. Unimplemented.");
	            }
	            else
	            {
	            	logger.info("The points of clip region are absolute locations");
	                // the points are absolute locations
	                if (path.getCbit())
	                {
	                    Path area = new Path();
	                	
	                	int j = 0;
	                	while(j < path.getPoints().length) 
	                	{
	                	    Path polygon = new Path();
	                	    boolean firstPoint = true;
	                	    
		                	for(; j < path.getPoints().length; ++j)
		                	{	
		                	    if ( firstPoint )
		                	    {
		                	        polygon.moveTo( path.getPoints()[j].x, path.getPoints()[j].y );
		                	        firstPoint = false;
		                	    }
		                	    else
		                	    {
		                	        polygon.lineTo( path.getPoints()[j].x, path.getPoints()[j].y );
		                	    }
		                	    
		                		if(path.getPointTypes()[j].getPointFlags() == PathPointTypeFlags.PathPointTypeCloseSubpath) 
		                		{
		                			++j;
		                			break;
		                		}
		                	}
		                	area.addPath( polygon );
	                	}
	                	area.close();
	                	graphics.clipPath( area, Op.REPLACE );
	                }
	                else
	                {
	                	logger.info("Locations are specified by EmfPlusPointF objects. Unimplemented.");
	                }
	            }
				break;
			default:
				logger.error("Regions of this type are not supported. Type: " + Integer.toHexString(node.getType()));
				break;
			}
		}
	}
	
	private void drawImage(EmfPlusDrawImage drawImage, EmfPlusImage imageObject, /*Graphics*/ Canvas graphics)
	{
		logger.info("Draw Image");
		switch(imageObject.getImageType())
		{
		case ImageDataTypeEnumeration.ImageDataTypeBitmap:
			logger.info("Image data type is Bitmap");
			EmfPlusBitmap bitmap = (EmfPlusBitmap) imageObject;
			try 
			{
			    InputStream in = new ByteArrayInputStream(bitmap.getImageData());
			    Bitmap decompressed = BitmapFactory.decodeStream( in );
			    
			    logger.debug( "Decompressed secondary order: color " + (Integer.toHexString( decompressed.getPixel( 0, 0 ) )) );
			    logger.debug( "Decompressed secondary order: bitmap size=" + drawImage.getDestRect().x + " " + drawImage.getDestRect().y + " " + drawImage.getDestRect().width + " " + drawImage.getDestRect().height );
			    
			    graphics.drawBitmap( decompressed, 
			            new Rect(0, 0, (int)drawImage.getDestRect().width, (int)drawImage.getDestRect().height ), 
			            new RectF(
			                    drawImage.getDestRect().x, 
			                    drawImage.getDestRect().y, 
			                    drawImage.getDestRect().x +
			                    drawImage.getDestRect().width,
			                    drawImage.getDestRect().y + 
			                    drawImage.getDestRect().height), 
			            null );
			} 
			catch (Exception e) 
			{
			    logger.error("Error drawing image: " + e.getMessage());
            }
			break;
		default:
			logger.info("Unknown Image Data Type: " + Integer.toHexString(imageObject.getImageType()));
			break;
		}
	}
}