/*
 *  Licensed to the Apache Software Foundation (ASF) under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
/**
 * @author Igor V. Stolyarov
 * @version $Revision$
 */

package quest.androidport;

import java.math.BigInteger;

/**
 * The Class IndexColorModel represents a color model in which the color values
 * of the pixels are read from a palette.
 * 
 * @since Android 1.0
 */
public class ColorModel {

    /**
     * The color map.
     */
    private int colorMap[]; // Color Map

    /**
     * The map size.
     */
    private int mapSize; // Color Map size

    /**
     * The transparent index.
     */
    private int transparentIndex; // Index of fully transparent pixel

    /**
     * The gray palette.
     */
    private boolean grayPalette; // Color Model has Color Map with Gray Pallete

    /**
     * The valid bits.
     */
    private BigInteger validBits; // Specify valid Color Map values

    /**
     * The Constant CACHESIZE.
     */
    private static final int CACHESIZE = 20; // Cache size. Cache used for

    // improving performace of selection
    // nearest color in Color Map

    /**
     * The cachetable.
     */
    private final int cachetable[] = new int[CACHESIZE * 2]; // Cache table -

    // used for

    // storing RGB values and that appropriate indices
    // in the Color Map

    /**
     * The next insert idx.
     */
    private int nextInsertIdx = 0; // Next index for insertion into Cache table

    /**
     * The total inserted.
     */
    private int totalInserted = 0; // Number of inserted values into Cache table

    /**
     * Instantiates a new index color model by building the color map from
     * arrays of red, green, and blue values.
     * 
     * @param bits
     *            the array of component masks.
     * @param size
     *            the size of the color map.
     * @param r
     *            the array giving the red components of the entries in the
     *            color map.
     * @param g
     *            the array giving the green components of the entries in the
     *            color map.
     * @param b
     *            the array giving the blue components of the entries in the
     *            color map.
     * @throws IllegalArgumentException
     *             if the size of the color map is less than one.
     * @throws ArrayIndexOutOfBoundsException
     *             if the size of one of the component arrays is less than the
     *             size of the color map.
     */
    public ColorModel(int bits, int size, byte r[], byte g[], byte b[]) 
    {
        createColorMap(size, r, g, b, null, -1);
        checkPalette();
    }

    /**
     * Gets the valid pixels.
     * 
     * @return the valid pixels.
     */
    public BigInteger getValidPixels() {
        return validBits;
    }

    /**
     * Gets the color map.
     * 
     * @param rgb
     *            the destination array where the color map is written.
     */
    public final void getRGBs(int rgb[]) {
        System.arraycopy(colorMap, 0, rgb, 0, mapSize);
    }

    /**
     * Gets the red component of the color map.
     * 
     * @param r
     *            the destination array.
     */
    public final void getReds(byte r[]) {
        for (int i = 0; i < mapSize; i++) {
            r[i] = (byte)(colorMap[i] >> 16);
        }
    }

    /**
     * Gets the green component of the color map.
     * 
     * @param g
     *            the destination array.
     */
    public final void getGreens(byte g[]) {
        for (int i = 0; i < mapSize; i++) {
            g[i] = (byte)(colorMap[i] >> 8);
        }
    }

    /**
     * Gets the blue component of the color map.
     * 
     * @param b
     *            the destination array.
     */
    public final void getBlues(byte b[]) {
        for (int i = 0; i < mapSize; i++) {
            b[i] = (byte)colorMap[i];
        }
    }

    /**
     * Gets the alpha component of the color map.
     * 
     * @param a
     *            the destination array.
     */
    public final void getAlphas(byte a[]) {
        for (int i = 0; i < mapSize; i++) {
            a[i] = (byte)(colorMap[i] >> 24);
        }
    }

    /**
     * Checks if the specified pixel is valid for this color model.
     * 
     * @param pixel
     *            the pixel.
     * @return true, if the pixel is valid.
     */
    public boolean isValid(int pixel) {
        if (validBits == null) {
            return (pixel >= 0 && pixel < mapSize);
        }
        return (pixel < mapSize && validBits.testBit(pixel));
    }

    public final int getRed(int pixel) {
        return (colorMap[pixel] >> 16) & 0xff;
    }

    public final int getRGB(int pixel) {
        return colorMap[pixel];
    }

    public final int getGreen(int pixel) {
        return (colorMap[pixel] >> 8) & 0xff;
    }

    
    public final int getBlue(int pixel) {
        return colorMap[pixel] & 0xff;
    }

    
    public final int getAlpha(int pixel) {
        return (colorMap[pixel] >> 24) & 0xff;
    }

    
    /**
     * Checks if this color model validates pixels.
     * 
     * @return true, if all pixels are valid, otherwise false.
     */
    public boolean isValid() {
        return (validBits == null);
    }

    
    public void finalize() {
        // TODO: implement
        return;
    }

    /**
     * Gets the index that represents the transparent pixel.
     * 
     * @return the index that represents the transparent pixel.
     */
    public final int getTransparentPixel() {
        return transparentIndex;
    }

    
    public int getTransparency() 
    {
        return 0xFF000000;
    }

    /**
     * Gets the size of the color map.
     * 
     * @return the map size.
     */
    public final int getMapSize() {
        return mapSize;
    }

    /**
     * Creates the color map.
     * 
     * @param size
     *            the size.
     * @param r
     *            the r.
     * @param g
     *            the g.
     * @param b
     *            the b.
     * @param a
     *            the a.
     * @param trans
     *            the trans.
     */
    private void createColorMap(int size, byte r[], byte g[], byte b[], byte a[], int trans) 
    {
        if (size < 1) 
        {
            return;
        }

        mapSize = size;
        colorMap = new int[mapSize];
        int alpha = 0;

        for (int i = 0; i < mapSize; i++) 
        {
            colorMap[i] = ((r[i] & 0xff) << 16) | ((g[i] & 0xff) << 8) | (b[i] & 0xff);

            if (trans == i) 
            {
                continue;
            }

            if (a == null) 
            {
                colorMap[i] |= 0xff000000;
            } 
            else 
            {
                alpha = a[i] & 0xff;
                if (alpha == 0xff) 
                {
                    colorMap[i] |= 0xff000000;
                } 
                else if (alpha == 0) 
                {
                    if (transparentIndex < 0) 
                    {
                        transparentIndex = i;
                    }
                } 
                else 
                {
                    colorMap[i] |= (a[i] & 0xff) << 24;
                }
            }

        }

    }

    /**
     * This method checking, if Color Map has Gray palette.
     */
    private void checkPalette() {
        grayPalette = false;
        int rgb = 0;

        for (int i = 0; i < mapSize; i++) {
            rgb = colorMap[i];
            if (((rgb >> 16) & 0xff) != ((rgb >> 8) & 0xff) || ((rgb >> 8) & 0xff) != (rgb & 0xff)) {
                return;
            }
        }
        grayPalette = true;
    }

    /**
     * Creates the bits.
     * 
     * @param hasAlpha
     *            the has alpha.
     * @return the int[].
     */
    private static int[] createBits(boolean hasAlpha) {

        int numChannels;
        if (hasAlpha) {
            numChannels = 4;
        } else {
            numChannels = 3;
        }

        int bits[] = new int[numChannels];
        for (int i = 0; i < numChannels; i++) {
            bits[i] = 8;
        }

        return bits;

    }

    /**
     * Checks if is gray palette.
     * 
     * @return true, if is gray palette.
     */
    boolean isGrayPallete() {
        return grayPalette;
    }

} 