
package quest.android.appportal.connection;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import quest.android.appportal.R;
import quest.android.appportal.common.ICOConverter;
import quest.android.appportal.common.SSLUtil;
import quest.android.appportal.common.Server;
import quest.android.appportal.common.Common;

import android.app.Activity;
import android.graphics.Bitmap;

import com.provisionnetworks.common.PNCipher;

/**
 * @author dens
 *
 */
public class PNClient
{
    private Logger logger = Logger.getLogger( PNClient.class.getName() );

    public static final String  HTTP_PROTO  = "http";
    public static final String  HTTPS_PROTO = "https";

    private static final int    REQUEST_TIMEOUT     = 60000;
    
    private List<String> serverList = null;
    
    private int                 tcpPort     = 8080;
    private boolean             useSSL      = false;
    
    private String              domain      = null;
    private String              username    = null;
    private String              password    = null;
    private String              clientIPAddress = null;
    private String              clientName  = null;    
    private boolean             useAlternateIPAddress = false;

    private String              activeCookie    = "";
    private String              activeServer    = ""; 

    private Activity activity = null;
    
    public PNClient( Activity activity, boolean useSSL, List<String> servers, int port,
            String username, String password, String domain )
            throws IOException
    {
        this.activity = activity;
        
        this.useSSL = useSSL;
        this.serverList = servers;
        this.tcpPort = port;
        this.username = username;
        this.password = password;
        this.domain = domain;
        this.clientIPAddress = Common.getLocalIpAddress();
        this.clientName = Common.getLocalHostName();

        if ( useSSL )
        {
            SSLUtil.disableCertChecking();
        }
    }

    public PNClient( Activity activity, String protocol, List<String> servers, int port, String username,
            String password, String domain ) throws IOException
    {
        this( activity, protocol.equalsIgnoreCase( HTTPS_PROTO ), servers, port, username, password, domain );
    }
    
    public PNClient( Activity activity, List<String> servers, int port, String username,
            String password, String domain ) throws IOException
    {
        this( activity, false, servers, port, username, password, domain );
    }

    private static String getValue( String str, String key )
    {
        return getValue( str, key, 0 );
    }

    private static String getValue( String str, String key, int start )
    {
        if ( str == null || key == null )
        {
            return null;
        }

        String val = null;
        int ind1 = str.indexOf( "<" + key + ">", start );
        int ind2 = str.indexOf( "</" + key + ">", ind1 );
        if ( ind1 > -1 && ind2 > -1 )
        {
            val = str.substring( ind1 + key.length() + 2, ind2 );
        }
        return val;
    }

    private String sendRequest( String server, String request, boolean ignoreErrors )
            throws IOException, PNException
    {
        try
        {
            String response = Common.EMTPY_STRING;
    
            URL url = new URL( (this.useSSL ? HTTPS_PROTO : HTTP_PROTO ) + "://" + server + ":"
                    + this.tcpPort );
            HttpURLConnection connection = (HttpURLConnection)url.openConnection();
    
            connection.setConnectTimeout( REQUEST_TIMEOUT );
            connection.setReadTimeout( REQUEST_TIMEOUT );
            connection.setDoOutput( true );
            connection.setRequestMethod( "POST" );
            connection.setRequestProperty( "User-Agent", "PNClient" );
            connection.setRequestProperty( "Accept", "*/*" );
            connection.setRequestProperty( "Content-Type", "text/xml" );
            
            connection.connect();
    
            OutputStream os = connection.getOutputStream();
            os.write( request.getBytes() );
            os.flush();
            InputStream is = connection.getInputStream();
    
            byte buffer[] = new byte[8192];
    
            int read = is.read( buffer );
    
            while ( read >= 0 )
            {
                response += new String( buffer, 0, read );
                read = is.read( buffer );
            }
    
            os.close();
            is.close();
            
            logger.debug( "Broker response: " + response );
    
            if ( ignoreErrors )
            {
                return response;
            }
    
            if ( !Common.isNullOrEmpty( response ) )
            {
                int errorCode = 0;
                String errorText = null;
                String errorCodeStr = getValue( response, "errorcode" );
    
                if ( errorCodeStr != null )
                {
                    errorCode = Common.parseIntSafe( errorCodeStr, 0 );
                    if ( errorCode != 0 )
                    {
                        errorText = getValue( response, "errortext" );
                        logger.error( "Broker Authentication error: " + errorText
                                + ". Code: " + errorCode );
                        throw new PNException( errorText, errorCode );
                    }
                }
                else
                {
                    logger.warn( "Unknown response: " + response );
                    
                    errorText = activity.getString( R.string.broker_error_unknown_response );
                    logger.error( "Broker Authentication error: " + errorText );
    
                    throw new PNException( errorText, errorCode );
                }
            }
            else
            {
                String errorText = activity.getString( R.string.broker_error_server_response_is_empty );
                logger.warn( errorText );
    
                throw new PNException( errorText, 0 );
            }
            
            return response;
        }
        catch( PNException e )
        {
            throw e;
        }
        catch ( Exception e )
        {
            return null;
        }
    }

    /**
     * Authenticates with the broker.
     * 
     * @throws IOException
     * @throws PNException
     */
    public void authenticate( ) throws IOException, PNException
    {
        logger.debug( "Authentication started" );

        byte[] bin = PNCipher.encode256( this.password );
        PNCipher.encrypt( bin );
        String encryptedPassword = PNCipher.binToHex( bin );
        String request = "<authenticate-req>"
                + "<domain>" + domain + "</domain>"
                + "<username>" + this.username + "</username>" 
                + "<password>" + encryptedPassword + "</password>"
                + "<clientname>" + this.clientName + "</clientname>"
                + "<clientipaddress>" + this.clientIPAddress + "</clientipaddress>" 
                + "<clientversion>" + "6.0" + "</clientversion>"
                + "<clientbitsperpixel>32</clientbitsperpixel>"
                + "<usealternateipaddress>" + (this.useAlternateIPAddress ? 1 : 0) + "</usealternateipaddress>"
                + "</authenticate-req>\n";

        String response = null;
        
        int i = 0;
        
        for ( ; i < serverList.size() && Common.isNullOrEmpty( response ) ; ++i )
        {
            response = sendRequest( serverList.get( i ), request, false );
        }
        --i;
        
        if ( Common.isNullOrEmpty( response ) )
        {
            throw new PNException( activity.getString( R.string.broker_error_no_logon_servers ), PNException.ERROR_NO_LOGON_SERVERS );
        }
        
        this.activeServer = serverList.get( i );
        
        activeCookie = getValue( response, "cookie" );
    }

    /**
     * Get application list for provided user
     * 
     * @return PNApplication array list
     * 
     * @throws IOException
     * @throws PNException
     */
    public ArrayList<PNApplication> getApplicationList() throws IOException,
            PNException
    {
        authenticate();

        String request = "<getapplicationlist-req><cookie>" + activeCookie
                + "</cookie></getapplicationlist-req>\n";
        String response = sendRequest( this.activeServer, request, false );

        String applicationTag = "<application>";

        ArrayList<PNApplication> appList = new ArrayList<PNApplication>();

        int ind = response.indexOf( applicationTag );

        while ( ind > -1 )
        {
            String name = getValue( response, "name", ind );
            String shortcut = getValue( response, "shortcut", ind );
            String path = getValue( response, "path", ind );
            String arguments = getValue( response, "arguments", ind );
            String workingdir = getValue( response, "workingdir", ind );
            String largeIconHex = getValue( response, "largeicon", ind ); // LargeIcon
            String smallIconHex = getValue( response, "smallicon", ind ); // SmallIcon
            String showcmd = getValue( response, "showcmd", ind );
            String foldername = getValue( response, "foldername", ind );

            PNApplication application = new PNApplication( name );

            ICOConverter converter = new ICOConverter();
            int[] rgb = converter.convertIcon( largeIconHex );
            Bitmap largeIcon = null;
            if ( rgb != null )
            {
                largeIcon = Bitmap.createBitmap( rgb, converter.getWidth(),
                        converter.getHeight(), Bitmap.Config.ARGB_8888 );
            }
            application.setLargeIcon( largeIcon ); // Largeicon

            converter = new ICOConverter();
            rgb = converter.convertIcon( smallIconHex );
            Bitmap smallIcon = null;
            if ( rgb != null )
            {
                smallIcon = Bitmap.createBitmap( rgb, converter.getWidth(),
                        converter.getHeight(), Bitmap.Config.ARGB_8888 );
            }
            application.setSmallIcon( smallIcon ); // Smallicon

            application.setShortcut( shortcut );
            application.setPath( path );
            application.setArguments( arguments );
            application.setWorkingDir( workingdir );
            application.setShowCmd( showcmd );
            application.setFolderName( foldername );
            appList.add( application );
            ind = response.indexOf( applicationTag, ind + 1 );
        }

        logger.debug( response );

        return appList;
    }

    /**
     * Get launch info for application
     * 
     * @param appName - application name
     * @return Special session info ( server name, address, application path etc.)
     *  
     * @throws IOException
     * @throws PNException
     */
    public PNSessionInfo launchApplication( String appName )
            throws IOException, PNException
    {
        authenticate();

        String request = "<launchapplication-req><cookie>" + activeCookie
                + "</cookie><appname>" + appName
                + "</appname></launchapplication-req>\n";
        String resp = sendRequest( this.activeServer, request, false );

        PNSessionInfo appInfo = new PNSessionInfo( appName );
        String serverName = getValue( resp, "servername" );
        String serverAddr = getValue( resp, "serveraddr" );
        String shortcut = getValue( resp, "shortcut" );
        String arguments = getValue( resp, "arguments" );
        String workingdir = getValue( resp, "workingdir" );
        String showcmd = getValue( resp, "showcmd" );
        String path = getValue( resp, "path" );
        String sessionSharing = getValue( resp, "sessionsharing" );
        String dnaProxy = getValue( resp, "dnaproxy" );

        appInfo.setServerName( serverName );
        appInfo.setServerAddr( serverAddr );
        appInfo.setShortcut( shortcut );
        appInfo.setArguments( arguments );
        appInfo.setWorkingdir( workingdir );
        appInfo.setShowcmd( showcmd );
        appInfo.setPath( path );
        appInfo.setSessionSharing( sessionSharing );
        appInfo.setDnaProxy( dnaProxy );
        
        return appInfo;
    }

    public PNFarmProperties getFarmProperties() throws IOException,
            PNException
    {
        authenticate();

        String request = "<getfarmproperties-req><cookie>" + activeCookie
                + "</cookie></getfarmproperties-req>";

        PNFarmProperties farmProps = new PNFarmProperties();

        String resp = sendRequest( this.activeServer, request, false );

        farmProps.setFarmName( getValue( resp, "farmname" ) );

        String serverTag = "<server>";
        int ind = resp.indexOf( serverTag );

        while ( ind > -1 )
        {
            String name = getValue( resp, "name", ind );
            String ip = getValue( resp, "ipaddress", ind );
            farmProps.addServer( name, ip );
            ind = resp.indexOf( serverTag, ind + 1 );
        }

        return farmProps;
    }

    public Server getBusyServer() throws IOException, PNException
    {
        authenticate();

        String request = "<getleastbusyserver-req><usealternateipaddress>0</usealternateipaddress></getleastbusyserver-req>\n";
        String resp = sendRequest( this.activeServer, request, false );

        String name = getValue( resp, "servername" );
        String ip = getValue( resp, "serveraddr" );

        Server busyServer = new Server( name, ip );

        return busyServer;
    }

    public String getUsername()
    {
        return username;
    }

    public void setUsername( String username )
    {
        this.activeCookie = null;
        this.username = username;
    }

    public String getPassword()
    {
        return password;
    }

    public void setPassword( String password )
    {
        this.activeCookie = null;
        this.password = password;
    }

    public String getDomain()
    {
        return domain;
    }

    public void setDomain( String domain )
    {
        this.activeCookie = null;
        this.domain = domain;
    }

    public boolean isUseSSL()
    {
        return useSSL;
    }

    public void setUseSSL( boolean useSSL )
    {
        this.activeCookie = null;
        this.useSSL = useSSL;
    }

    public String getServer()
    {
        return activeServer;
    }

    public int getPort()
    {
        return tcpPort;
    }
}
