
package quest.android.appportal.farm;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import quest.android.appportal.common.PropertyDefinitions;
import quest.android.appportal.common.Common;

public class PNLocation implements PropertyDefinitions
{
    private Logger logger = Logger.getLogger( PNLocation.class.getName() );
    
    // Static members
    public static final String      HTTP_PROTO      = "HTTP";
    public static final String      HTTPS_PROTO     = "HTTPS";

    public static final String      Name            = "name";
    public static final String      NAT             = "nat";
    public static final String      Protocol        = "protocol";
    public static final String      ProxyBypassList = "proxybypasslist";
    public static final String      ProxyServer     = "proxyserver";
    public static final String      RDPonSSL        = "rdponssl";
    public static final String      Servers         = "servers";
    public static final String      SSLGateway      = "sslgateway";
    public static final String      TCPPort         = "tcpport";
    public static final String      UseProxy        = "useproxy";
    public static final String      LocationUuid    = "locationuuid";

    // *****************************************************************************
    // Private location members
    // *****************************************************************************
    private String name = "";
    private Properties locationProps = new Properties();

    // *****************************************************************************

    public PNLocation( String locationName )
    {
        logger.debug( "Location [" + locationName + "] created." );
        
        this.setName( locationName );
        this.locationProps.put( Name, locationName );
        this.locationProps.put( NAT, FALSE );
        this.locationProps.put( Protocol, HTTP_PROTO );
        this.locationProps.put( ProxyBypassList, EMPTY );
        this.locationProps.put( ProxyServer, EMPTY );
        this.locationProps.put( RDPonSSL, FALSE );
        this.locationProps.put( Servers, EMPTY ); // add brokers in format
                                                  // "firstbroker;secondbroker;...;lastbroker"
        this.locationProps.put( SSLGateway, EMPTY );
        this.locationProps.put( TCPPort, "8080" );
        this.locationProps.put( UseProxy, FALSE );
        
        this.setUuid( Common.generateUUID() );
    }

    public String getUuid()
    {
        return getStringVal( LocationUuid );
    }
    
    public String getName()
    {
        return name;
    }

    private boolean getBoolVal( String key )
    {
        return this.locationProps.get( key ).equals( TRUE );
    }
    
    private void setBoolVal( String key, boolean value )
    {
        this.locationProps.put( key, value ? TRUE : FALSE );
    }

    private String getStringVal( String key )
    {
        return (String) this.locationProps.get( key );
    }
    
    private void setStringVal( String key, String value )
    {
        this.locationProps.put( key, value );
    }

    public boolean isNATEnabled()
    {
        return getBoolVal( NAT );
    }

    public String getProtocol()
    {
        return getStringVal( Protocol );
    }

    public boolean isHttpsProtocol()
    {
        return (getProtocol().equalsIgnoreCase( HTTPS_PROTO ));
    }
    
    public void setHttpsProtocol( boolean enabled )
    {
        setProtocol( enabled ? HTTPS_PROTO : HTTP_PROTO );
    }

    public String getProxyBypassList()
    {
        return getStringVal( ProxyBypassList );
    }

    public String getProxyServer()
    {
        return getStringVal( ProxyServer );
    }

    public boolean isRdpOverSslEnabled()
    {
        return getBoolVal( RDPonSSL );
    }

    public String getSslGateway()
    {
        return getStringVal( SSLGateway );
    }

    public String getTcpPort()
    {
        return getStringVal( TCPPort );
    }

    public boolean isProxyEnabled()
    {
        return getBoolVal( UseProxy );
    }

    public ArrayList<String> getServersList()
    {
        String servers = getStringVal( Servers );
        
        if ( Common.isNullOrEmpty( servers ) )
        {
            return new ArrayList<String>();
        }
        
        ArrayList<String> serverList = new ArrayList<String>(Arrays.asList( servers.split( ";" )));
        return serverList;
    }

    public void setName( String name )
    {
        if ( name == null )
        {
            name = "";
        }
        setStringVal( Name, name );
        this.name = name;
    }
    
    public void setNatEnabled( boolean enabled )
    {
        setBoolVal( NAT, enabled );
    }
    
    public void setProtocol( String protocol )
    {
        setStringVal( Protocol, protocol);
    }
    
    public void setProxyBypassList( String list )
    {
        setStringVal( ProxyBypassList, list);
    }
    
    public void setProxyServer( String proxyServer )
    {
        setStringVal( ProxyServer, proxyServer );
    }
    
    public void setRdpOverSslEnabled( boolean enabled )
    {
        setBoolVal( RDPonSSL, enabled);
    }

    public void addServer( String server )
    {
        String servers = getStringVal( Servers );
        setStringVal( Servers, Common.isNullOrEmpty( servers ) ? server : servers + ";" + server );
    }

    public void removeServer( String server ) throws PNLocationException
    {
        ArrayList<String> servers = new ArrayList<String>(Arrays.asList( getStringVal( Servers ).split( ";" ) ));
        if ( !servers.remove( server ) )
        {
            throw new PNLocationException("Cannot find server: " + server);
        }
        setServers(servers);
    }
    
    public void removeServer( int index ) throws PNLocationException
    {
        ArrayList<String> servers = new ArrayList<String>(Arrays.asList( getStringVal( Servers ).split( ";" ) ));
        if ( index < 0 || index >= servers.size() || servers.remove( index ) == null )
        {
            throw new PNLocationException("Cannot find server: " + index);
        }
        setServers(servers);
    }
    
    public void setServers( ArrayList<String> servers )
    {
        String res = EMPTY;
        
        for ( int i = 0; servers != null && i < servers.size(); ++i )
        {
            res += servers.get( i ) + ( i < servers.size() - 1 ? ";" : EMPTY );
        }
        
        setStringVal( Servers, res );
    }
    
    public void setSslGateway( String gateway )
    {
        setStringVal( SSLGateway, gateway );
    }
    
    public void setTcpPort( String port )
    {
        setStringVal( TCPPort, port );
    }

    public void setProxyUsingEnabled( boolean enabled )
    {
        setBoolVal( UseProxy, enabled );
    }
    
    public void setUuid( String guid )
    {
        this.setStringVal( LocationUuid, guid );
    }
    
    // *****************************************************************************
    // Utility functions
    // *****************************************************************************
    
    public PNLocation clone()
    {
        PNLocation clone = null;
        clone = new PNLocation( this.getName() );
        clone.locationProps = new Properties( this.locationProps );
        return clone; 
    }
    
    public void saveLocation( Document doc, Node root )
    {
        Enumeration<?> keys = this.locationProps.propertyNames();
        Element locNode = doc.createElement( "Location" );
        
        while ( keys.hasMoreElements() )
        {
            String param = (String)keys.nextElement();
            String value = this.locationProps.getProperty( param );
            locNode.setAttribute( param, value );
        }
        
        root.appendChild( locNode );
    }
    
    public static PNLocation loadLocation( Node root )
    {
        if ( !root.getNodeName().equals( "Location" ) )
        {
            return null;
        }
        
        String name = ((Element) root).getAttribute( Name );
        
        if ( name == null )
        {
            return null;
        }
        
        PNLocation loc = null;
        
        loc = new PNLocation( name );
        
        NamedNodeMap attrs = root.getAttributes();
        
        for ( int i = 0; i < attrs.getLength(); ++i )
        {
            String key = attrs.item( i ).getNodeName();
            String value = attrs.item( i ).getNodeValue();
            loc.locationProps.put( key, value );
        }        
        
        return loc;
    }
    
}
